# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""CLI-related utilities for Charmcraft."""
import datetime
import sys
from dataclasses import dataclass
from typing import Iterable, Optional

from craft_cli import emit

from charmcraft.env import is_charmcraft_running_in_managed_mode


class SingleOptionEnsurer:
    """Argparse helper to ensure that the option is specified only once, converting it properly.

    Receives a callable to convert the string from command line to the desired object.

    Example of use:

        parser.add_argument('-n', '--number',  type=SingleOptionEnsurer(int), required=True)

    No lower limit is checked, that is verified with required=True in the argparse definition.
    """

    def __init__(self, converter):
        self.converter = converter
        self.count = 0

    def __call__(self, value):
        """Run by argparse to validate and convert the given argument."""
        self.count += 1
        if self.count > 1:
            raise ValueError("the option can be specified only once")
        return self.converter(value)


@dataclass(frozen=True)
class ResourceOption:
    """Argparse helper to validate and convert a 'resource' option.

    Receives a callable to convert the string from command line to the desired object.

    Example of use:

        parser.add_argument('--resource',  type=ResourceOption())
    """

    name: Optional[str] = None
    revision: Optional[int] = None

    def __call__(self, value):
        """Run by argparse to validate and convert the given argument."""
        parts = [x.strip() for x in value.split(":")]
        parts = [p for p in parts if p]
        if len(parts) == 2:
            name, revision = parts
            try:
                revision = int(revision)
            except ValueError:
                pass
            else:
                if revision >= 0:
                    return ResourceOption(name, revision)
        msg = (
            "the resource format must be <name>:<revision> (revision being a non-negative integer)"
        )
        raise ValueError(msg)


def confirm_with_user(prompt, default=False) -> bool:
    """Query user for yes/no answer.

    If stdin is not a tty, the default value is returned.

    If user returns an empty answer, the default value is returned.

    :returns: True if answer starts with [yY], False if answer starts with [nN],
        otherwise the default.
    """
    if is_charmcraft_running_in_managed_mode():
        raise RuntimeError("confirmation not yet supported in managed-mode")

    if not sys.stdin.isatty():
        return default

    choices = " [Y/n]: " if default else " [y/N]: "

    with emit.pause():
        reply = input(prompt + choices).lower().strip()

    if reply and reply[0] == "y":
        return True
    elif reply and reply[0] == "n":
        return False
    else:
        return default


def humanize_list(items: Iterable[str], conjunction: str) -> str:
    """Format a list into a human-readable string.

    :param items: list to humanize, must not be empty
    :param conjunction: the conjunction used to join the final element to
                        the rest of the list (e.g. 'and').
    """
    if not items:
        raise ValueError("Cannot humanize an empty list.")
    *initials, final = map(repr, sorted(items))
    if not initials:
        return final
    return f"{', '.join(initials)} {conjunction} {final}"


def format_timestamp(dt: datetime.datetime) -> str:
    """Convert a datetime object (with or without timezone) to a string.

    The format is

        <DATE>T<TIME>Z

    Always in UTC.
    """
    if dt.tzinfo is not None and dt.tzinfo.utcoffset(None) is not None:
        # timezone aware
        dtz = dt.astimezone(datetime.timezone.utc)
    else:
        # timezone naive, assume it's UTC
        dtz = dt
    return dtz.strftime("%Y-%m-%dT%H:%M:%SZ")



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for CharmBuilder."""
import pathlib

import pytest
from pyfakefs.fake_filesystem import FakeFilesystem
from pytest_subprocess import FakeProcess

from charmcraft import charm_builder, errors, utils

pytestmark = [
    # Always use pyfakefs and pytest-subprocess
    pytest.mark.usefixtures("fs", "fake_process")
]

REQUIREMENTS_FILES = [pytest.param("", id="empty"), "ops~=2.5", "requests\nops"]


@pytest.fixture()
def builder(fs: FakeFilesystem) -> charm_builder.CharmBuilder:
    fs.cwd = "/root"
    fs.makedirs("build")
    fs.makedirs("install")
    charm_file = fs.create_file("src/charm.py")

    return charm_builder.CharmBuilder(
        builddir=pathlib.Path("build"),
        installdir=pathlib.Path("install"),
        entrypoint=charm_file.path,
        requirements=[pathlib.Path("requirements.txt")],
    )


@pytest.mark.usefixtures("fake_process")
@pytest.mark.parametrize(
    ("requirements", "python_packages", "binary_python_packages"),
    [
        pytest.param("", ["yolo"], [], id="empty-requirements"),
    ],
)
def test_install_strict_dependencies_validation_error(
    fs: FakeFilesystem, builder, requirements, python_packages, binary_python_packages
):
    fs.create_file("requirements.txt", contents=requirements)
    builder.strict_dependencies = True
    builder.python_packages = python_packages
    builder.binary_python_packages = binary_python_packages

    with pytest.raises(errors.DependencyError):
        builder._install_strict_dependencies("pip")


@pytest.mark.usefixtures("fake_process")
def test_install_strict_dependencies_no_requirement_paths(builder):
    builder.requirement_paths = []

    with pytest.raises(errors.DependencyError):
        builder._install_strict_dependencies("pip")


@pytest.mark.parametrize("requirements", REQUIREMENTS_FILES)
def test_install_strict_dependencies_pip_failure(
    fs, fake_process: FakeProcess, builder, requirements
):
    fs.create_file("requirements.txt", contents=requirements)
    no_binary_packages = utils.get_package_names(requirements.splitlines(keepends=False))
    no_binary_packages_str = ",".join(sorted(no_binary_packages))
    fake_process.register(
        [
            "/pip",
            "install",
            *([f"--no-binary={no_binary_packages_str}"] if no_binary_packages else []),
            "--requirement=requirements.txt",
        ],
        returncode=1,
    )

    with pytest.raises(RuntimeError):
        builder._install_strict_dependencies("/pip")


@pytest.mark.parametrize("requirements", REQUIREMENTS_FILES)
def test_install_strict_dependencies_success(
    fs: FakeFilesystem, fake_process: FakeProcess, builder, requirements
):
    fs.create_file("requirements.txt", contents=requirements)
    no_binary_packages = utils.get_package_names(requirements.splitlines(keepends=False))
    no_binary_packages_str = ",".join(sorted(no_binary_packages))
    fake_process.register(
        [
            "/pip",
            "install",
            *([f"--no-binary={no_binary_packages_str}"] if no_binary_packages else []),
            "--requirement=requirements.txt",
        ],
        returncode=0,
    )

    builder._install_strict_dependencies("/pip")



# Copyright 2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Tests for the instrumentator module."""

import json
from unittest.mock import patch

import pytest

from charmcraft import instrum
from charmcraft.instrum import Timer, _Measurements


@pytest.fixture()
def fake_times():
    """Provide times from 5 by 10."""
    with patch("charmcraft.instrum.time", side_effect=list(range(5, 100, 10))):
        yield


# -- tests for the Measurement class


def test_measurement_start_end_cycle(fake_times):
    """Basic start and end cycle."""
    measurements = _Measurements()

    mid = measurements.start("test msg", {"foo": "bar"})
    assert measurements.parents == [None, mid]
    recorded = measurements.measurements[mid]
    assert recorded == {
        "parent": None,
        "msg": "test msg",
        "extra": {"foo": "bar"},
        "tstart": 5,
        "tend": None,
    }

    measurements.end(mid)
    assert measurements.parents == [None]
    recorded = measurements.measurements[mid]
    assert recorded == {
        "parent": None,
        "msg": "test msg",
        "extra": {"foo": "bar"},
        "tstart": 5,
        "tend": 15,
    }


def test_measurement_tree():
    """A tree is built for the measurements."""
    measurements = _Measurements()

    # start first
    mid_1 = measurements.start("test msg", {})
    assert measurements.parents == [None, mid_1]
    assert measurements.measurements[mid_1]["parent"] is None

    # start second
    mid_2 = measurements.start("test msg", {})
    assert measurements.parents == [None, mid_1, mid_2]
    assert measurements.measurements[mid_2]["parent"] == mid_1

    # start third
    mid_3 = measurements.start("test msg", {})
    assert measurements.parents == [None, mid_1, mid_2, mid_3]
    assert measurements.measurements[mid_3]["parent"] == mid_2

    # end them all
    measurements.end(mid_3)
    assert measurements.parents == [None, mid_1, mid_2]
    measurements.end(mid_2)
    assert measurements.parents == [None, mid_1]
    measurements.end(mid_1)
    assert measurements.parents == [None]


def test_measurement_extra_info_complex():
    """No objects are leaked in the extra info."""
    measurements = _Measurements()

    weird_object = object()

    mid = measurements.start("test msg", {"foo": 42, "bar": weird_object})
    assert measurements.measurements[mid]["extra"] == {"foo": "42", "bar": str(weird_object)}


def test_measurement_overlapped_measurements():
    """The instrumentator set up overlapped measurements."""
    measurements = _Measurements()

    # start first and second
    mid_1 = measurements.start("test msg", {})
    measurements.start("test msg", {})

    # end first before second
    with pytest.raises(ValueError):
        measurements.end(mid_1)


def test_measurement_dump(tmp_path, fake_times):
    """Dump the measurements content to a dump file."""
    measurements = _Measurements()
    mid = measurements.start("test msg", {"foo": "bar"})
    measurements.end(mid)

    measures_filepath = tmp_path / "measures.json"
    with patch("charmcraft.instrum._baseline", 3):
        measurements.dump(measures_filepath)

    dumped_content = json.loads(measures_filepath.read_text())
    assert dumped_content == {
        "__meta__": {"baseline": 3},  # patched value
        mid: {
            "extra": {"foo": "bar"},
            "msg": "test msg",
            "parent": None,
            "tstart": 2,  # relative to baseline
            "tend": 12,  # relative to baseline
        },
    }


def test_measurement_merge_simple(tmp_path, fake_times):
    """Merge a simple measurement into current structure."""
    measures_filepath = tmp_path / "measures.json"

    # start the outer measure
    measurements_outer = _Measurements()
    mid_outer = measurements_outer.start("outer msg", {})

    # meanwhile, start the inner measure
    measurements_inner = _Measurements()
    mid_inner = measurements_inner.start("inner msg", {})
    measurements_inner.end(mid_inner)
    with patch("charmcraft.instrum._baseline", 12):
        measurements_inner.dump(measures_filepath)

    # confirm dumped info
    dumped_content = json.loads(measures_filepath.read_text())
    assert dumped_content == {
        "__meta__": {"baseline": 12},  # patched value
        mid_inner: {
            "extra": {},
            "msg": "inner msg",
            "parent": None,
            "tstart": 3,  # relative to baseline
            "tend": 13,  # relative to baseline
        },
    }

    # merge from it and check merged structure
    measurements_outer.merge_from(measures_filepath)
    merged = measurements_outer.measurements[mid_inner]
    assert merged["parent"] == mid_outer  # not None anymore
    assert merged["tstart"] == 15  # back to absolute
    assert merged["tend"] == 25  # back to absolute


def test_measurement_merge_complex(tmp_path, fake_times):
    """Merge a complex measurements tree into current structure."""
    measures_filepath = tmp_path / "measures.json"

    # start two outer measures
    measurements_outer = _Measurements()
    measurements_outer.start("outer msg 1", {})  # fake time: 5
    mid_outer_2 = measurements_outer.start("outer msg 2", {})  # fake time: 15

    # in the inner measurement start two root ones, one with a "child"
    measurements_inner = _Measurements()
    mid_inner_1 = measurements_inner.start("inner msg 1", {})  # fake time: 25
    mid_inner_2 = measurements_inner.start("inner msg 2", {})  # fake time: 35
    measurements_inner.end(mid_inner_2)  # fake time: 45
    measurements_inner.end(mid_inner_1)  # fake time: 55
    mid_inner_3 = measurements_inner.start("inner msg 3", {})  # fake time: 65
    measurements_inner.end(mid_inner_3)  # fake time: 75

    # dump it all
    with patch("charmcraft.instrum._baseline", 20):
        measurements_inner.dump(measures_filepath)

    # confirm dumped info
    dumped_content = json.loads(measures_filepath.read_text())
    assert dumped_content == {
        "__meta__": {"baseline": 20},  # patched value
        mid_inner_1: {
            "extra": {},
            "msg": "inner msg 1",
            "parent": None,
            "tstart": 5,  # relative to baseline
            "tend": 35,  # relative to baseline
        },
        mid_inner_2: {
            "extra": {},
            "msg": "inner msg 2",
            "parent": mid_inner_1,
            "tstart": 15,  # relative to baseline
            "tend": 25,  # relative to baseline
        },
        mid_inner_3: {
            "extra": {},
            "msg": "inner msg 3",
            "parent": None,
            "tstart": 45,  # relative to baseline
            "tend": 55,  # relative to baseline
        },
    }

    # merge from it and check merged structure
    measurements_outer.merge_from(measures_filepath)
    merged_1 = measurements_outer.measurements[mid_inner_1]
    assert merged_1["parent"] == mid_outer_2  # the parent is the "current" outer measure
    assert merged_1["tstart"] == 25  # back to absolute
    assert merged_1["tend"] == 55  # back to absolute
    merged_2 = measurements_outer.measurements[mid_inner_2]
    assert merged_2["parent"] == mid_inner_1  # inside the structure, parent respected
    assert merged_2["tstart"] == 35  # back to absolute
    assert merged_2["tend"] == 45  # back to absolute
    merged_3 = measurements_outer.measurements[mid_inner_3]
    assert merged_3["parent"] == mid_outer_2  # the parent is the "current" outer measure
    assert merged_3["tstart"] == 65  # back to absolute
    assert merged_3["tend"] == 75  # back to absolute


# -- tests for the Timer class


def test_timer_as_context_manager(fake_times, monkeypatch):
    """Use test as a context manager."""
    measurements = _Measurements()
    monkeypatch.setattr(instrum, "_measurements", measurements)
    assert measurements.measurements == {}

    with Timer("test message", foo=42):
        pass

    (recorded,) = measurements.measurements.values()
    assert recorded == {
        "parent": None,
        "msg": "test message",
        "extra": {"foo": "42"},
        "tstart": 5,
        "tend": 15,
    }


def test_timer_as_context_manager_with_mark(fake_times, monkeypatch):
    """Use test as a context manager, hitting marks in the code block."""
    measurements = _Measurements()
    monkeypatch.setattr(instrum, "_measurements", measurements)

    @Timer("test message", foo=42)
    def test_function(a, b):
        assert a == 17
        assert b == 35

    assert measurements.measurements == {}
    test_function(17, b=35)

    (recorded,) = measurements.measurements.values()
    assert recorded == {
        "parent": None,
        "msg": "test message",
        "extra": {"foo": "42"},
        "tstart": 5,
        "tend": 15,
    }


def test_timer_as_decorator(fake_times, monkeypatch):
    """Use test as a decorator."""
    measurements = _Measurements()
    monkeypatch.setattr(instrum, "_measurements", measurements)
    assert measurements.measurements == {}

    with Timer("test message", foo=42) as timer:
        timer.mark("middle 1")
        timer.mark("middle 2")

    (recorded_1, recorded_2, recorded_3) = measurements.measurements.values()

    # first we have the context manager itself, that starts at the very
    # beginning and ends on the first mark
    assert recorded_1 == {
        "parent": None,
        "msg": "test message",
        "extra": {"foo": "42"},
        "tstart": 5,
        "tend": 15,
    }

    # then the measure between both marks
    assert recorded_2 == {
        "parent": None,
        "msg": "middle 1",
        "extra": {},
        "tstart": 25,
        "tend": 35,
    }

    # finally the measure between second mark and the end of the context manager
    assert recorded_3 == {
        "parent": None,
        "msg": "middle 2",
        "extra": {},
        "tstart": 45,
        "tend": 55,
    }



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Main entry point module for all the tool functionality."""

import logging
import os
import sys

import craft_providers.errors
import craft_store.errors
from craft_cli import (
    ArgumentParsingError,
    CommandGroup,
    CraftError,
    Dispatcher,
    EmitterMode,
    GlobalArgument,
    ProvideHelpException,
    emit,
)

from charmcraft import __version__, config, env, utils
from charmcraft.commands import analyze, clean, extensions, init, pack, store, version
from charmcraft.commands.store.client import ALTERNATE_AUTH_ENV_VAR
from charmcraft.const import SHARED_CACHE_ENV_VAR
from charmcraft.parts import setup_parts

# set up all the libs' loggers in DEBUG level so their content is grabbed by craft-cli's Emitter
for lib_name in ("craft_providers", "craft_parts", "craft_store"):
    logger = logging.getLogger(lib_name)
    logger.setLevel(logging.DEBUG)


# the summary of the whole program
GENERAL_SUMMARY = """
Charmcraft helps build, package and publish operators on Charmhub.

Together with the Python Operator Framework, charmcraft simplifies
operator development and collaboration.

See https://charmhub.io/publishing for more information.
"""

# Collect commands in different groups, for easier human consumption. Note that this is not
# declared in each command because it's much easier to do this separation/grouping in one
# central place and not distributed in several classes/files. Also note that order here is
# important when listing commands and showing help.
_basic_commands = [
    analyze.AnalyzeCommand,
    clean.CleanCommand,
    pack.PackCommand,
    init.InitCommand,
    version.VersionCommand,
]
_charmhub_commands = [
    # auth
    store.LoginCommand,
    store.LogoutCommand,
    store.WhoamiCommand,
    # name handling
    store.RegisterCharmNameCommand,
    store.RegisterBundleNameCommand,
    store.UnregisterNameCommand,
    store.ListNamesCommand,
    # pushing files and checking revisions
    store.UploadCommand,
    store.ListRevisionsCommand,
    # release process, and show status
    store.ReleaseCommand,
    store.PromoteBundleCommand,
    store.StatusCommand,
    store.CloseCommand,
    # libraries support
    store.CreateLibCommand,
    store.PublishLibCommand,
    store.ListLibCommand,
    store.FetchLibCommand,
    # resources support
    store.ListResourcesCommand,
    store.UploadResourceCommand,
    store.ListResourceRevisionsCommand,
]
_extensions_commands = [
    extensions.ExtensionsCommand,
    extensions.ListExtensionsCommand,
    extensions.ExpandExtensionsCommand,
]
COMMAND_GROUPS = [
    CommandGroup("Basic", _basic_commands),
    CommandGroup("Charmhub", _charmhub_commands),
    CommandGroup("Extensions", _extensions_commands),
]

# non-charmcraft useful environment variables to log
EXTRA_ENVIRONMENT = ("DESKTOP_SESSION", "XDG_CURRENT_DESKTOP", SHARED_CACHE_ENV_VAR)


def _get_system_details():
    """Produce details about the system."""
    # prepare the useful environment variables: all CHARMCRAFT* (hiding AUTH keys)
    # and desktop/session
    useful_env = {
        name: value
        for name, value in os.environ.items()
        if name.startswith("CHARMCRAFT") or name in EXTRA_ENVIRONMENT
    }
    if ALTERNATE_AUTH_ENV_VAR in useful_env:
        useful_env[ALTERNATE_AUTH_ENV_VAR] = "<hidden>"
    env_string = ", ".join(f"{name}={value!r}" for name, value in sorted(useful_env.items()))
    if not env_string:
        env_string = "None"

    os_platform = utils.get_os_platform()
    return f"System details: {os_platform}; Environment: {env_string}"


def _emit_error(error, cause=None):
    """Emit the error in a centralized way so we can alter it consistently."""
    # set the cause, if any
    if cause is not None:
        error.__cause__ = cause

    # if it's a charmcraft running inside a provided instance, do not report the internal logpath
    if env.is_charmcraft_running_in_managed_mode():
        error.logpath_report = False

    # finally, emit
    emit.error(error)


def main(argv=None):
    """Provide the main entry point."""
    if env.is_charmcraft_running_in_managed_mode():
        logpath = env.get_managed_environment_log_path()
    else:
        logpath = None

    emit.init(
        EmitterMode.BRIEF,
        "charmcraft",
        f"Starting charmcraft version {__version__}",
        log_filepath=logpath,
    )

    if argv is None:
        argv = sys.argv

    extra_global_options = [
        GlobalArgument(
            "project_dir",
            "option",
            "-p",
            "--project-dir",
            "Specify the project's directory (defaults to current)",
        ),
    ]

    # process
    try:
        setup_parts()

        # load the dispatcher and put everything in motion
        dispatcher = Dispatcher(
            "charmcraft",
            COMMAND_GROUPS,
            summary=GENERAL_SUMMARY,
            extra_global_args=extra_global_options,
        )
        global_args = dispatcher.pre_parse_args(argv[1:])
        loaded_config = config.load(global_args["project_dir"])
        dispatcher.load_command(loaded_config)
        emit.debug(_get_system_details())
        retcode = dispatcher.run()

    except ArgumentParsingError as err:
        print(err, file=sys.stderr)  # to stderr, as argparse normally does
        emit.ended_ok()
        retcode = 1
    except ProvideHelpException as err:
        print(err, file=sys.stderr)  # to stderr, as argparse normally does
        emit.ended_ok()
        retcode = 0
    except CraftError as err:
        _emit_error(err)
        retcode = err.retcode
    except craft_store.errors.CraftStoreError as err:
        error = CraftError(f"craft-store error: {err}")
        _emit_error(error)
        retcode = 1
    except craft_providers.errors.ProviderError as err:
        _emit_error(CraftError(err.brief, details=err.details, resolution=err.resolution))
        retcode = 1
    except KeyboardInterrupt as exc:
        error = CraftError("Interrupted.")
        _emit_error(error, cause=exc)
        retcode = 1
    except Exception as err:
        error = CraftError(f"charmcraft internal error: {err!r}")
        _emit_error(error, cause=err)
        retcode = 1
    else:
        emit.ended_ok()
        if retcode is None:
            retcode = 0

    return retcode


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""The charm package builder.

This is a standalone script run by the Charm Plugin.
"""

import argparse
import errno
import hashlib
import os
import pathlib
import shutil
import subprocess
import sys
from typing import List

from charmcraft import instrum
from charmcraft.const import (
    DEPENDENCIES_HASH_FILENAME,
    DISPATCH_CONTENT,
    DISPATCH_FILENAME,
    HOOKS_DIRNAME,
    MANDATORY_HOOK_NAMES,
    STAGING_VENV_DIRNAME,
    VENV_DIRNAME,
)
from charmcraft.env import get_charm_builder_metrics_path
from charmcraft.errors import DependencyError
from charmcraft.jujuignore import JujuIgnore, default_juju_ignore
from charmcraft.utils import (
    collect_charmlib_pydeps,
    get_pip_command,
    get_pip_version,
    get_requirements_file_package_names,
    make_executable,
    validate_strict_dependencies,
)

MINIMUM_PIP_VERSION = (23, 0)
KNOWN_GOOD_PIP_URL = "https://files.pythonhosted.org/packages/ba/19/e63fb4e0d20e48bd2167bb7e857abc0e21679e24805ba921a224df8977c0/pip-23.2.1.tar.gz"


def relativise(src, dst):
    """Build a relative path from src to dst."""
    return pathlib.Path(os.path.relpath(str(dst), str(src.parent)))


class CharmBuilder:
    """The package builder."""

    def __init__(
        self,
        builddir: pathlib.Path,
        installdir: pathlib.Path,
        entrypoint: pathlib.Path,
        allow_pip_binary: bool = None,
        binary_python_packages: List[str] = None,
        python_packages: List[str] = None,
        requirements: List[pathlib.Path] = None,
        strict_dependencies: bool = False,
    ) -> None:
        self.builddir = builddir
        self.installdir = installdir
        self.entrypoint = entrypoint
        self.allow_pip_binary = allow_pip_binary
        self.binary_python_packages = binary_python_packages
        self.python_packages = python_packages
        self.requirement_paths = requirements
        self.strict_dependencies = strict_dependencies
        self.ignore_rules = self._load_juju_ignore()
        self.ignore_rules.extend_patterns([f"/{STAGING_VENV_DIRNAME}"])

        self.charmlib_deps = collect_charmlib_pydeps(builddir)
        print("Collected charmlib dependencies:", self.charmlib_deps)

    def build_charm(self) -> None:
        """Build the charm."""
        print(f"Building charm in {str(self.installdir)!r}")

        if self.installdir.exists():
            shutil.rmtree(str(self.installdir))
        self.installdir.mkdir()

        linked_entrypoint = self.handle_generic_paths()
        self.handle_dispatcher(linked_entrypoint)
        self.handle_dependencies()

    def _load_juju_ignore(self):
        ignore = JujuIgnore(default_juju_ignore)
        path = self.builddir / ".jujuignore"
        if path.exists():
            with path.open("r", encoding="utf-8") as ignores:
                ignore.extend_patterns(ignores)
        return ignore

    def create_symlink(self, src_path, dest_path):
        """Create a symlink in dest_path pointing relatively like src_path.

        It also verifies that the linked dir or file is inside the project.
        """
        resolved_path = src_path.resolve()
        if self.builddir in resolved_path.parents:
            relative_link = relativise(src_path, resolved_path)
            dest_path.symlink_to(relative_link)
        else:
            rel_path = src_path.relative_to(self.builddir)
            print(f"Ignoring symlink because targets outside the project: {str(rel_path)!r}")

    @instrum.Timer("Handling generic paths")
    def handle_generic_paths(self):
        """Handle all files and dirs except what's ignored and what will be handled later.

        Works differently for the different file types:
        - regular files: hard links
        - directories: created
        - symlinks: respected if are internal to the project
        - other types (blocks, mount points, etc): ignored
        """
        print("Linking in generic paths")

        for basedir, dirnames, filenames in os.walk(str(self.builddir), followlinks=False):
            abs_basedir = pathlib.Path(basedir)
            rel_basedir = abs_basedir.relative_to(self.builddir)

            # process the directories
            ignored = []
            for pos, name in enumerate(dirnames):
                rel_path = rel_basedir / name
                abs_path = abs_basedir / name

                if self.ignore_rules.match(str(rel_path), is_dir=True):
                    print(f"Ignoring directory because of rules: {str(rel_path)!r}")
                    ignored.append(pos)
                elif abs_path.is_symlink():
                    dest_path = self.installdir / rel_path
                    self.create_symlink(abs_path, dest_path)
                else:
                    dest_path = self.installdir / rel_path
                    dest_path.mkdir(mode=abs_path.stat().st_mode)

            # in the future don't go inside ignored directories
            for pos in reversed(ignored):
                del dirnames[pos]

            # process the files
            for name in filenames:
                rel_path = rel_basedir / name
                abs_path = abs_basedir / name

                if self.ignore_rules.match(str(rel_path), is_dir=False):
                    print(f"Ignoring file because of rules: {str(rel_path)!r}")
                elif abs_path.is_symlink():
                    dest_path = self.installdir / rel_path
                    self.create_symlink(abs_path, dest_path)
                elif abs_path.is_file():
                    dest_path = self.installdir / rel_path
                    try:
                        os.link(str(abs_path), str(dest_path))
                    except PermissionError:
                        # when not allowed to create hard links
                        shutil.copy2(str(abs_path), str(dest_path))
                    except OSError as e:
                        if e.errno != errno.EXDEV:
                            raise
                        shutil.copy2(str(abs_path), str(dest_path))
                else:
                    print(f"Ignoring file because of type: {str(rel_path)!r}")

        # the linked entrypoint is calculated here because it's when it's really in the build dir
        return self.installdir / self.entrypoint.relative_to(self.builddir)

    @instrum.Timer("Handling dispatcher")
    def handle_dispatcher(self, linked_entrypoint):
        """Handle modern and classic dispatch mechanisms."""
        # dispatch mechanism, create one if wasn't provided by the project
        dispatch_path = self.installdir / DISPATCH_FILENAME
        if not dispatch_path.exists():
            print("Creating the dispatch mechanism")
            dispatch_content = DISPATCH_CONTENT.format(
                entrypoint_relative_path=linked_entrypoint.relative_to(self.installdir)
            )
            with dispatch_path.open("wt", encoding="utf8") as fh:
                fh.write(dispatch_content)
                make_executable(fh)

        # bunch of symlinks, to support old juju: verify that any of the already included hooks
        # in the directory is not linking directly to the entrypoint, and also check all the
        # mandatory ones are present
        dest_hookpath = self.installdir / HOOKS_DIRNAME
        if not dest_hookpath.exists():
            dest_hookpath.mkdir()

        # get those built hooks that we need to replace because they are pointing to the
        # entrypoint directly and we need to fix the environment in the middle
        current_hooks_to_replace = []
        for node in dest_hookpath.iterdir():
            if node.resolve() == linked_entrypoint:
                current_hooks_to_replace.append(node)
                node.unlink()
                print(f"Replacing existing hook {node.name!r} as it's a symlink to the entrypoint")

        # include the mandatory ones and those we need to replace
        hooknames = MANDATORY_HOOK_NAMES | {x.name for x in current_hooks_to_replace}
        for hookname in hooknames:
            print(f"Creating the {hookname!r} hook script pointing to dispatch")
            dest_hook = dest_hookpath / hookname
            if not dest_hook.exists():
                relative_link = relativise(dest_hook, dispatch_path)
                dest_hook.symlink_to(relative_link)

    def _calculate_dependencies_hash(self):
        """Calculate a hash for all current dependencies."""
        all_deps = []
        for req_file in self.requirement_paths:
            all_deps.append(req_file.read_text())
        all_deps.extend(self.binary_python_packages)
        all_deps.extend(self.python_packages)
        all_deps.extend(self.charmlib_deps)
        deps_mashup = "".join(map(repr, all_deps))
        return hashlib.sha1(deps_mashup.encode("utf8")).hexdigest()

    @instrum.Timer("Installing dependencies")
    def _install_dependencies(self, staging_venv_dir):
        """Install all dependencies in a specific directory."""
        # create virtualenv using the host environment python
        with instrum.Timer("Creating venv"):
            _process_run(["python3", "-m", "venv", str(staging_venv_dir)])
        pip_cmd = str(_find_venv_bin(staging_venv_dir, "pip"))

        with instrum.Timer("Ensuring a recent enough pip..."):
            # pip 20 (included with focal) has dependency resolution issues related to
            # common charm dependencies (e.g. ops). Resolve this by updating to a
            # known working version of pip.
            if get_pip_version(pip_cmd) < MINIMUM_PIP_VERSION:
                _process_run([pip_cmd, "install", f"pip@{KNOWN_GOOD_PIP_URL}"])

        with instrum.Timer("Installing all dependencies"):
            if self.strict_dependencies:
                self._install_strict_dependencies(pip_cmd)
                return

            # Legacy non-strict dependencies.
            # This method is not valid for any bases added after 2024-01-01 or for DEVEL bases.
            try:
                _process_run(
                    get_pip_command(
                        [pip_cmd, "install"],
                        self.requirement_paths,
                        source_deps=[*self.python_packages, *self.charmlib_deps],
                        binary_deps=self.binary_python_packages,
                    )
                )
            except RuntimeError:
                print(
                    "WARNING: Initial package installation failed. "
                    "Falling back to older method, which may leave your charm "
                    "in an un-runnable state."
                )
                if self.binary_python_packages:
                    # install python packages, allowing binary packages
                    cmd = [pip_cmd, "install", "--upgrade"]  # base command
                    cmd.extend(self.binary_python_packages)  # the python packages to install
                    _process_run(cmd)
                if self.python_packages:
                    # install python packages from source
                    cmd = [pip_cmd, "install", "--upgrade", "--no-binary", ":all:"]  # base command
                    cmd.extend(self.python_packages)  # the python packages to install
                    _process_run(cmd)
                if self.requirement_paths:
                    # install dependencies from requirement files
                    cmd = [pip_cmd, "install", "--upgrade", "--no-binary", ":all:"]  # base command
                    for reqspath in self.requirement_paths:
                        cmd.append(f"--requirement={reqspath}")  # the dependencies file(s)
                    _process_run(cmd)
                if self.charmlib_deps:
                    # install charmlibs python dependencies
                    cmd = [pip_cmd, "install", "--upgrade", "--no-binary", ":all:"]  # base command
                    cmd.extend(self.charmlib_deps)  # the python packages to install
                _process_run(cmd)

    def _install_strict_dependencies(self, pip_cmd: str) -> None:
        if not self.requirement_paths:
            raise DependencyError(
                "No requirements files have been passed to the charm builder.",
                details="Strict dependencies mode needs at least one requirements file.",
            )
        validate_strict_dependencies(
            get_requirements_file_package_names(*self.requirement_paths),
            self.charmlib_deps,
            self.python_packages or [],
            self.binary_python_packages or [],
        )
        _process_run(
            get_pip_command(
                [pip_cmd, "install"],
                self.requirement_paths,
                binary_deps=self.binary_python_packages or [],
            )
        )

    def handle_dependencies(self):
        """Handle from-directory and virtualenv dependencies."""
        print("Handling dependencies")
        if not (
            self.requirement_paths
            or self.binary_python_packages
            or self.python_packages
            or self.charmlib_deps
        ):
            print("No dependencies to handle")
            return

        staging_venv_dir = self.builddir / STAGING_VENV_DIRNAME
        hash_file = self.builddir / DEPENDENCIES_HASH_FILENAME

        # find out if current dependencies are the same than the last run.
        current_deps_hash = self._calculate_dependencies_hash()
        print(f"Current dependencies hash: {current_deps_hash!r}")
        if not staging_venv_dir.exists():
            print("Dependencies directory not found")
            same_dependencies = False
        elif hash_file.exists():
            try:
                previous_deps_hash = hash_file.read_text(encoding="utf8")
            except Exception as exc:
                print(f"Problems reading the dependencies hash file: {exc}")
                same_dependencies = False
            else:
                print(f"Previous dependencies hash: {previous_deps_hash!r}")
                same_dependencies = previous_deps_hash == current_deps_hash
        else:
            print("Dependencies hash file not found")
            same_dependencies = False

        if same_dependencies:
            print("Reusing installed dependencies, they are equal to last run ones")
        else:
            print("Installing dependencies")
            self._install_dependencies(staging_venv_dir)

            # save the hash file after all successful installations
            hash_file.write_text(current_deps_hash, encoding="utf8")

        # always copy the virtualvenv site-packages directory to /venv in charm
        basedir = pathlib.Path(STAGING_VENV_DIRNAME)
        site_packages_dir = _find_venv_site_packages(basedir)
        shutil.copytree(site_packages_dir, self.installdir / VENV_DIRNAME)


def _find_venv_bin(basedir: pathlib.Path, exec_base: str) -> pathlib.Path:
    """Determine the venv executable in different platforms."""
    if sys.platform == "win32":
        return basedir / "Scripts" / f"{exec_base}.exe"

    return basedir / "bin" / exec_base


def _find_venv_site_packages(basedir):
    """Determine the venv site-packages directory in different platforms."""
    output = subprocess.check_output(
        ["python3", "-c", "import sys; v=sys.version_info; print(f'{v.major} {v.minor}')"],
        text=True,
    )
    major, minor = output.strip().split(" ")

    if sys.platform == "win32":
        return basedir / f"Python{major}{minor}" / "site-packages"

    return basedir / "lib" / f"python{major}.{minor}" / "site-packages"


def _process_run(cmd: List[str]) -> None:
    """Run an external command logging its output.

    :raises CraftError: if execution crashes or ends with return code not zero.
    """
    print(f"Running external command {cmd}")
    try:
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
    except Exception as exc:
        raise RuntimeError(f"Subprocess command {cmd} execution crashed: {exc!r}")

    for line in proc.stdout:
        print(f"   :: {line.rstrip()}")
    retcode = proc.wait()

    if retcode:
        raise RuntimeError(f"Subprocess command {cmd} execution failed with retcode {retcode}")


def _parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--entrypoint",
        default="src/charm.py",
        type=pathlib.Path,
        help="The charm entry point. Default is 'src/charm.py'.",
    )
    parser.add_argument(
        "--builddir",
        default=".",
        type=pathlib.Path,
        help="The charm source directory. Default is current.",
    )
    parser.add_argument(
        "--installdir",
        required=True,
        type=pathlib.Path,
        help="The build destination directory",
    )
    parser.add_argument(
        "-b",
        "--binary-package",
        action="append",
        help="Binary Python package to install before requirements.",
    )
    strict_group = parser.add_mutually_exclusive_group()
    strict_group.add_argument(
        "-p",
        "--package",
        action="append",
        help="Python package to install before requirements.",
    )
    strict_group.add_argument(
        "--strict-dependencies", action="store_true", help="Use strict dependencies."
    )
    parser.add_argument(
        "-r",
        "--requirement",
        action="append",
        type=pathlib.Path,
        help="Requirements file to install dependencies from.",
    )

    return parser.parse_args()


def main():
    """Run the command-line interface."""
    options = _parse_arguments()

    print("Starting charm builder")
    builder = CharmBuilder(
        builddir=options.builddir,
        installdir=options.installdir,
        entrypoint=options.entrypoint,
        binary_python_packages=options.binary_package or [],
        python_packages=options.package or [],
        requirements=options.requirement or [],
        strict_dependencies=options.strict_dependencies,
    )
    builder.build_charm()


if __name__ == "__main__":
    with instrum.Timer("Full charm_builder.py main"):
        main()
    instrum.dump(get_charm_builder_metrics_path())



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charmcraft error classes."""
import io
import pathlib
from typing import Iterable, Mapping

from craft_cli import CraftError


class BadLibraryPathError(CraftError):
    """Subclass to provide a specific error for a bad library path."""

    def __init__(self, path):
        super().__init__(
            "Charm library path {} must conform to lib/charms/<charm>/vN/<libname>.py"
            "".format(path)
        )


class BadLibraryNameError(CraftError):
    """Subclass to provide a specific error for a bad library name."""

    def __init__(self, name):
        super().__init__(
            f"Charm library name {name!r} must conform to charms.<charm>.vN.<libname>"
        )


class InvalidCharmPathError(CraftError):
    """The path provided is not the source directory for a valid charm."""

    def __init__(self, path: pathlib.Path):
        super().__init__(
            f"Path does not contain source for a valid charm: {path}",
            resolution=(
                "Ensure the given path is a directory containing valid charmcraft.yaml "
                "and metadata.yaml files."
            ),
        )


class DuplicateCharmsError(CraftError):
    """Duplicate charms were found on disk for the same name.

    If source is True, this refers to charm sources. Otherwise, it refers to files.
    """

    _sources_resolution = (
        "Remove duplicate charms or specify directories with --include-charm. "
        "These can be seen with --verbosity=debug"
    )
    _files_resolution = (
        "Ensure each charm generates only one output file. "
        "Files can be seen with --verbosity=debug"
    )

    def __init__(self, charms: Mapping[str, Iterable[pathlib.Path]], source: bool = True):
        import charmcraft.utils

        charm_names = charmcraft.utils.humanize_list(charms.keys(), "and")
        super().__init__(
            f"Duplicate charms found: {charm_names}",
            details=self._format_details(charms),
            resolution=self._sources_resolution if source else self._files_resolution,
            logpath_report=False,
            reportable=False,
        )

    @staticmethod
    def _format_details(charms: Mapping[str, Iterable[pathlib.Path]]) -> str:
        # 6 is the length of "CHARM", the left side of the header.
        longest_name = max([max(len(k) for k in charms), 5])
        path_tree_line_format = "{name:>" + str(longest_name) + "} : {path}"
        details = io.StringIO()
        print("Charms with duplicate paths:", file=details)
        print(path_tree_line_format.format(name="CHARM", path="PATHS"), file=details)
        for charm, paths in charms.items():
            path_iter = iter(paths)
            print(path_tree_line_format.format(name=charm, path=next(path_iter)), file=details)
            for path in path_iter:
                print(path_tree_line_format.format(name="", path=path), file=details)
        return details.getvalue()


class DependencyError(CraftError):
    """Errors related to dependencies."""


class InvalidDependenciesError(DependencyError):
    """In strict dependencies mode, some binary dependencies."""


class MissingDependenciesError(DependencyError):
    """In strict dependencies mode, some dependencies are missing from requirements files."""

    def __init__(self, extra_dependencies: Iterable[str]):
        self.extra_dependencies = sorted(extra_dependencies)
        extra_deps_str = ", ".join(self.extra_dependencies)
        super().__init__(
            "Some dependencies were not included in requirements files.",
            details=f"Missing dependencies: {extra_deps_str}",
            resolution="Ensure all missing dependencies are included in a requirements file.",
            # TODO: Docs URL
            reportable=False,
        )


class ExtensionError(CraftError):
    """Error related to extension handling."""



# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'pack' command."""
import argparse
import pathlib
from typing import Dict, List

import yaml
from craft_cli import ArgumentParsingError, CraftError, emit

from charmcraft import env, instrum, package
from charmcraft.cmdbase import BaseCommand
from charmcraft.utils import find_charm_sources, get_charm_name_from_path, load_yaml

# the minimum set of files in a bundle
MANDATORY_FILES = ["bundle.yaml", "README.md"]

_overview = """
Build and pack a charm operator package or a bundle.

You can `juju deploy` the resulting `.charm` or bundle's `.zip`
file directly, or upload it to Charmhub with `charmcraft upload`.

For the charm you must be inside a charm directory with a valid
`metadata.yaml`, `requirements.txt` including the `ops` package
for the Python operator framework, and an operator entrypoint,
usually `src/charm.py`.  See `charmcraft init` to create a
template charm directory structure.

For the bundle you must already have a `bundle.yaml` (can be
generated by Juju) and a README.md file.
"""


class PackCommand(BaseCommand):
    """Build the bundle or the charm.

    It uses the 'type' key in the configuration to decide which.
    """

    name = "pack"
    help_msg = "Build the charm or bundle"
    overview = _overview
    common = True

    def fill_parser(self, parser):
        """Add own parameters to the general parser."""
        self.include_format_option(parser)
        parser.add_argument(
            "--debug",
            action="store_true",
            help="Launch shell in build environment upon failure",
        )
        parser.add_argument(
            "--destructive-mode",
            action="store_true",
            help=(
                "Pack charm using current host which may result in breaking "
                "changes to system configuration"
            ),
        )
        parser.add_argument(
            "--shell",
            action="store_true",
            help="Launch shell in build environment in lieu of packing",
        )
        parser.add_argument(
            "--shell-after",
            action="store_true",
            help="Launch shell in build environment after packing",
        )
        parser.add_argument(
            "--bases-index",
            action="append",
            type=int,
            help="Index of 'bases' configuration to build (can be used multiple "
            "times); zero-based, defaults to all",
        )
        parser.add_argument(
            "--force",
            action="store_true",
            help="Force packing even after finding lint errors",
        )
        parser.add_argument(
            "--measure",
            type=pathlib.Path,
            help="Dump measurements to the specified file",
        )
        include_charm_group = parser.add_mutually_exclusive_group()
        include_charm_group.add_argument(
            "--include-all-charms",
            action="store_true",
            help="For bundles, pack all charms whose source is inside the bundle directory",
        )
        include_charm_group.add_argument(
            "--include-charm",
            action="append",
            type=pathlib.Path,
            help="For bundles, pack the charm in the referenced path. Can be used multiple times",
        )
        parser.add_argument(
            "--output-bundle",
            type=pathlib.Path,
            help="Write the bundle configuration to this path",
        )

    def run(self, parsed_args: argparse.Namespace) -> None:
        """Run the command."""
        self._check_config(config_file=True)

        builder = package.Builder(
            config=self.config,
            force=parsed_args.force,
            debug=parsed_args.debug,
            shell=parsed_args.shell,
            shell_after=parsed_args.shell_after,
            measure=parsed_args.measure,
        )

        # decide if this will work on a charm or a bundle
        if self.config.type == "charm":
            if parsed_args.include_all_charms:
                raise ArgumentParsingError(
                    "--include-all-charms can only be used when packing a bundle. "
                    f"Currently trying to pack: {self.config.project.dirpath}"
                )
            if parsed_args.include_charm:
                raise ArgumentParsingError(
                    "--include-charm can only be used when packing a bundle. "
                    f"Currently trying to pack: {self.config.project.dirpath}"
                )
            if parsed_args.output_bundle:
                raise ArgumentParsingError(
                    "--output-bundle can only be used when packing a bundle. "
                    f"Currently trying to pack: {self.config.project.dirpath}"
                )
            self._check_config(bases=True)
            with instrum.Timer("Whole pack run"):
                self._pack_charm(parsed_args, builder)
        elif self.config.type == "bundle":
            if parsed_args.shell:
                package.launch_shell()
                return
            bundle_filepath = self.config.project.dirpath / "bundle.yaml"
            bundle = load_yaml(bundle_filepath)
            if bundle is None:
                raise CraftError(f"Missing or invalid main bundle file: {str(bundle_filepath)!r}.")
            if parsed_args.include_all_charms:
                charm_names = bundle.get("applications", {}).keys()
                charms = find_charm_sources(self.config.project.dirpath, charm_names)
            elif parsed_args.include_charm:
                charms: Dict[str, pathlib.Path] = {}
                for path in parsed_args.include_charm:
                    if not path.is_absolute():
                        path = self.config.project.dirpath / path
                    name = get_charm_name_from_path(path)
                    charms[name] = path
            else:
                charms = {}
            with instrum.Timer("Whole pack run"):
                self._pack_bundle(parsed_args, charms, builder)
            if parsed_args.output_bundle:
                with parsed_args.output_bundle.open("wt") as file:
                    yaml.safe_dump(bundle, file)
        else:
            raise CraftError(f"Unknown type {self.config.type!r} in charmcraft.yaml")

        if parsed_args.measure:
            instrum.dump(parsed_args.measure)

    def _validate_bases_indices(self, bases_indices):
        """Validate that bases index is valid."""
        if bases_indices is None:
            return

        msg = "Bases index '{}' is invalid (must be >= 0 and fit in configured bases)."
        len_configured_bases = len(self.config.bases)
        for bases_index in bases_indices:
            if bases_index < 0:
                raise CraftError(msg.format(bases_index))
            if bases_index >= len_configured_bases:
                raise CraftError(msg.format(bases_index))

    def _pack_charm(self, parsed_args, builder: package.Builder) -> List[pathlib.Path]:
        """Pack a charm."""
        self._validate_bases_indices(parsed_args.bases_index)

        # build
        emit.progress("Packing the charm.")
        charms = builder.run(
            parsed_args.bases_index,
            destructive_mode=parsed_args.destructive_mode,
        )

        # avoid showing results when run inside a container (the outer charmcraft
        # is responsible of the final message to the user)
        if env.is_charmcraft_running_in_managed_mode():
            return

        if parsed_args.format:
            info = {"charms": charms}
            emit.message(self.format_content(parsed_args.format, info))
        else:
            emit.message("Charms packed:")
            for charm in charms:
                emit.message(f"    {charm}")

    def _pack_bundle(
        self,
        parsed_args: argparse.Namespace,
        charms: Dict[str, pathlib.Path],
        builder: package.Builder,
        overwrite_bundle: bool = False,
    ) -> None:
        """Pack a bundle."""
        emit.progress("Packing the bundle.")
        project = self.config.project

        if self.config.parts:
            config_parts = self.config.parts.copy()
        else:
            # "parts" not declared, create an implicit "bundle" part
            config_parts = {"bundle": {"plugin": "bundle"}}

        # a part named "bundle" using plugin "bundle" is special and has
        # predefined values set automatically.
        bundle_part = config_parts.get("bundle")
        if bundle_part and bundle_part.get("plugin") == "bundle":
            # set prime filters
            for fname in MANDATORY_FILES:
                fpath = project.dirpath / fname
                if not fpath.exists():
                    raise CraftError(f"Missing mandatory file: {str(fpath)!r}.")
            prime = bundle_part.setdefault("prime", [])
            prime.extend(MANDATORY_FILES)

            # set source if empty or not declared in charm part
            if not bundle_part.get("source"):
                bundle_part["source"] = str(project.dirpath)

        # run the parts lifecycle
        emit.debug(f"Parts definition: {config_parts}")

        try:
            output_files = builder.pack_bundle(
                charms=charms,
                base_indeces=parsed_args.bases_index or [],
                destructive_mode=parsed_args.destructive_mode,
                overwrite=overwrite_bundle,
            )
        except (RuntimeError, CraftError) as error:
            if parsed_args.debug:
                emit.debug(f"Error when running PRIME step: {error}")
                package.launch_shell()
            raise

        if parsed_args.format:
            info = {"bundles": [str(b) for b in output_files.bundles]}
            if output_files.charms:
                info["charms"] = [str(c) for c in output_files.charms]
            emit.message(self.format_content(parsed_args.format, info))
        else:
            emit.message(f"Created {str(output_files.bundles[0])!r}.")

        if parsed_args.shell_after:
            package.launch_shell()



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'clean' command."""

from craft_cli import emit

from charmcraft import providers
from charmcraft.cmdbase import BaseCommand
from charmcraft.utils import get_host_architecture

_overview = """
Purge Charmcraft project's artifacts, including:

- LXD Containers created for building charm(s)
- Multipass Containers created for building charm(s)
"""


class CleanCommand(BaseCommand):
    """Clean project artifacts."""

    name = "clean"
    help_msg = "Purge project artifacts"
    overview = _overview
    common = True

    def run(self, parsed_args):
        """Run the clean command.

        First, a build plan is created.
        Then each item in the build plan is cleaned.
        """
        self._check_config(config_file=True, bases=True)
        project_path = self.config.project.dirpath
        emit.message(f"Cleaning project {self.config.name!r}.")
        provider = providers.get_provider()
        build_plan = providers.create_build_plan(
            bases=self.config.bases,
            bases_indices=None,
            destructive_mode=False,
            managed_mode=False,
            provider=provider,
        )

        for plan in build_plan:
            instance_name = providers.get_instance_name(
                project_name=self.config.name,
                project_path=project_path,
                bases_index=plan.bases_index,
                build_on_index=plan.build_on_index,
                target_arch=get_host_architecture(),
            )

            emit.debug(f"Cleaning environment {instance_name!r}")
            provider.clean_project_environments(instance_name=instance_name)

        emit.message(f"Cleaned project {self.config.name!r}.")



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Craft store-related generic utilities."""
import enum
import functools
from dataclasses import dataclass
from typing import Optional

from craft_cli import CraftError


@functools.total_ordering
@enum.unique
class Risk(enum.Enum):
    """Standard risk tracks for a channel, orderable but not comparable to an int."""

    STABLE = 0
    CANDIDATE = 1
    BETA = 2
    EDGE = 3

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self.value == other.value
        return NotImplemented


@dataclass(frozen=True)
class ChannelData:
    """Data class for a craft store channel."""

    track: Optional[str]
    risk: Risk
    branch: Optional[str]

    @classmethod
    def from_str(cls, name: str):
        """Parse a channel name from a string using the standard store semantics.

        https://snapcraft.io/docs/channels
        """
        invalid_channel_error = CraftError(f"Invalid channel name: {name!r}")
        parts = name.split("/")
        track: Optional[str] = None
        branch: Optional[str] = None
        if len(parts) == 1:  # Just the risk, e.g. "stable"
            try:
                risk = Risk[parts[0].upper()]
            except KeyError:
                raise invalid_channel_error from None
        elif len(parts) == 2:
            try:  # risk/branch, e.g. "stable/debug"
                risk = Risk[parts[0].upper()]
                branch = parts[1]
            except KeyError:
                try:  # track/risk, e.g. "latest/stable"
                    risk = Risk[parts[1].upper()]
                    track = parts[0]
                except KeyError:
                    raise invalid_channel_error from None
        elif len(parts) == 3:  # Fully defined, e.g. "latest/stable/debug"
            try:
                risk = Risk[parts[1].upper()]
            except KeyError:
                raise invalid_channel_error from None
            track, _, branch = parts
        else:
            raise invalid_channel_error
        return cls(track, risk, branch)

    @property
    def name(self) -> str:
        """Get the channel name as a string."""
        risk = self.risk.name.lower()
        return "/".join(i for i in (self.track, risk, self.branch) if i is not None)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Platform-related Charmcraft utilities."""
import dataclasses
import pathlib
import platform

import distro


@dataclasses.dataclass(frozen=True)
class OSPlatform:
    """Description of an operating system platform."""

    system: str
    release: str
    machine: str


# translations from what the platform module informs to the term deb and
# snaps actually use
ARCH_TRANSLATIONS = {
    "aarch64": "arm64",
    "armv7l": "armhf",
    "i686": "i386",
    "ppc": "powerpc",
    "ppc64le": "ppc64el",
    "x86_64": "amd64",
    "AMD64": "amd64",  # Windows support
}


def get_os_platform(filepath=pathlib.Path("/etc/os-release")):
    """Determine a system/release combo for an OS using /etc/os-release if available."""
    system = platform.system()
    release = platform.release()
    machine = platform.machine()

    if system == "Linux":
        info = distro.info()
        system = info.get("id", system)
        # Treat Ubuntu derivatives as Ubuntu, as they should be compatible.
        if system != "ubuntu" and "ubuntu" in info.get("like", "").split():
            system = "ubuntu"
        release = info.get("version", release)

    return OSPlatform(system=system, release=release, machine=machine)


def get_host_architecture():
    """Get host architecture in deb format suitable for base definition."""
    os_platform = get_os_platform()
    return ARCH_TRANSLATIONS.get(os_platform.machine, os_platform.machine)



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'init' command."""

import os
import re
from datetime import date
from typing import Optional

from craft_cli import CraftError, emit

from charmcraft.cmdbase import BaseCommand
from charmcraft.utils import get_templates_environment, make_executable

try:
    import pwd
except ImportError:
    pwd = None  # type: ignore[assignment]

# the available profiles and in which directory the template can be found
PROFILES = {
    "simple": "init-simple",
    "kubernetes": "init-kubernetes",
    "machine": "init-machine",
}
DEFAULT_PROFILE = "simple"


_overview = """
Initialize a charm operator package tree and files.

This command will modify the directory to create the necessary files for a
charm operator package. By default it will work in the current directory.

Available profiles are:
    simple:
        A basic kubernetes charm with lot of texts helping the developer
        to navigate their first charm by following the instructions.

    kubernetes:
        A basic Kubernetes charm with example container.

    machine:
        A basic charm but meant to be deployed in machine-based environments,
        without container requirements.

Depending on the profile choice, Charmcraft will setup the following tree of
files and directories:

    .
    ├── charmcraft.yaml        - Charm build configuration
    ├── CONTRIBUTING.md        - Instructions for how to build and develop
    │                             your charm
    ├── LICENSE                - Your charm license, we recommend Apache 2
    ├── pyproject.toml         - Configuration for testing, formatting and
    │                             linting tools
    ├── README.md              - Frontpage for your charmhub.io/charm/
    ├── requirements.txt       - PyPI dependencies for your charm, with `ops`
    ├── src
    │   └── charm.py           - Minimal operator using Python operator framework
    ├── tests
    │   ├── integration
    │   │   └── test_charm.py  - Integration tests
    │   └── unit
    │       └── test_charm.py  - Unit tests
    └── tox.ini                - Configuration for tox, the tool to run all tests

You will need to edit at least charmcraft.yaml and README.md.

Your minimal operator code is in src/charm.py which uses the Python operator
framework from https://github.com/canonical/operator and there are some
example unit and integration tests with a harness to run them.
"""

SUCCESS_MESSAGE = """\
Charmed operator package file and directory tree initialised.

Now edit the following package files to provide fundamental charm metadata
and other information:

charmcraft.yaml
src/charm.py
README.md
"""


def _get_users_full_name_gecos() -> Optional[str]:
    """Get user's full name from Gecos (/etc/passwd)."""
    try:
        return pwd.getpwuid(os.getuid()).pw_gecos.split(",", 1)[0]
    except KeyError:
        return None


class InitCommand(BaseCommand):
    """Initialize a directory to be a charm project."""

    name = "init"
    help_msg = "Initialize a charm operator package tree and files"
    overview = _overview
    common = True

    def fill_parser(self, parser):
        """Specify command's specific parameters."""
        parser.add_argument("--name", help="The name of the charm; defaults to the directory name")
        parser.add_argument(
            "--author",
            help="The charm author; defaults to the current user name per GECOS",
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Initialize even if the directory is not empty (will not overwrite files)",
        )
        parser.add_argument(
            "--profile",
            choices=list(PROFILES),
            default=DEFAULT_PROFILE,
            help=f"Use the specified project profile (defaults to '{DEFAULT_PROFILE}')",
        )

    def run(self, args):
        """Execute command's actual functionality."""
        init_dirpath = self.config.project.dirpath
        if not init_dirpath.exists():
            init_dirpath.mkdir(parents=True)
        elif any(init_dirpath.iterdir()) and not args.force:
            tpl = "{!r} is not empty (consider using --force to work on nonempty directories)"
            raise CraftError(tpl.format(str(init_dirpath)))
        emit.debug(f"Using project directory {str(init_dirpath)!r}")

        if args.author is None and pwd is not None:
            args.author = _get_users_full_name_gecos()

        if not args.author:
            raise CraftError(
                "Unable to automatically determine author's name, specify it with --author"
            )

        if not args.name:
            args.name = init_dirpath.name
            emit.debug(f"Set project name to '{args.name}'")

        if not re.match(r"[a-z][a-z0-9-]*[a-z0-9]$", args.name):
            raise CraftError(
                f"{args.name} is not a valid charm name. "
                "The name must start with a lowercase letter "
                "and contain only alphanumeric characters and hyphens."
            )

        context = {
            "name": args.name,
            "author": args.author,
            "year": date.today().year,
            "class_name": "".join(re.split(r"\W+", args.name.title())) + "Charm",
        }

        template_directory = PROFILES[args.profile]
        env = get_templates_environment(template_directory)

        executables = ["run_tests", "src/charm.py"]
        for template_name in env.list_templates():
            if not template_name.endswith(".j2"):
                continue
            template = env.get_template(template_name)
            template_name = template_name[:-3]
            emit.debug(f"Rendering {template_name}")
            path = init_dirpath / template_name
            if path.exists():
                continue
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("wt", encoding="utf8") as fh:
                out = template.render(context)
                fh.write(out)
                if template_name in executables and os.name == "posix":
                    make_executable(fh)
                    emit.debug("  made executable")
        for line in SUCCESS_MESSAGE.split("\n"):
            emit.message(line)



# Copyright 2021-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Craft-parts setup, lifecycle and plugins."""

from typing import Any, Dict

from craft_parts import plugins
from craft_parts.parts import PartSpec

from charmcraft.parts.bundle import BundlePlugin
from charmcraft.parts.charm import CharmPlugin, CharmPluginProperties
from charmcraft.parts.lifecycle import PartsLifecycle
from charmcraft.parts.reactive import ReactivePlugin, ReactivePluginProperties

__all__ = [
    "CharmPlugin",
    "CharmPluginProperties",
    "ReactivePlugin",
    "ReactivePluginProperties",
    "setup_parts",
    "process_part_config",
    "PartsLifecycle",
]


def setup_parts():
    """Initialize craft-parts plugins."""
    plugins.register({"charm": CharmPlugin, "bundle": BundlePlugin, "reactive": ReactivePlugin})


def process_part_config(data: Dict[str, Any]) -> Dict[str, Any]:
    """Validate and fill the given part data against/with common and plugin models.

    :param data: The part data to use.

    :return: The part data validated and completed with plugin defaults.
    """
    if not isinstance(data, dict):
        raise TypeError("value must be a dictionary")

    # copy the original data, we'll modify it
    spec = data.copy()

    plugin_name = spec.get("plugin")
    if not plugin_name:
        raise ValueError("'plugin' not defined")

    plugin_class = plugins.get_plugin_class(plugin_name)

    # validate plugin properties
    plugin_properties = plugin_class.properties_class.unmarshal(spec)

    # validate common part properties
    part_spec = plugins.extract_part_properties(spec, plugin_name=plugin_name)
    PartSpec(**part_spec)

    # get plugin properties data if it's model based (otherwise it's empty), and
    # update with the received config
    if isinstance(plugin_properties, plugins.PluginModel):
        full_config = plugin_properties.dict(by_alias=True)
    else:
        full_config = {}
    full_config.update(data)

    return full_config



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft project handle config.yaml file."""

import contextlib
import logging
import pathlib
import shutil
from typing import TYPE_CHECKING, Optional

import pydantic
import yaml
from craft_cli import CraftError, emit

from charmcraft.const import JUJU_CONFIG_FILENAME
from charmcraft.format import format_pydantic_errors
from charmcraft.metafiles import read_yaml
from charmcraft.models.config import JujuConfig

if TYPE_CHECKING:
    from charmcraft.models.charmcraft import CharmcraftConfig

logger = logging.getLogger(__name__)


def parse_config_yaml(charm_dir: pathlib.Path, allow_broken=False) -> Optional[JujuConfig]:
    """Parse project's config.yaml.

    :param charm_dir: Directory to read config.yaml from.

    :returns: a JujuConfig object.

    :raises: CraftError if config.yaml is not valid.
    """
    try:
        config = read_yaml(charm_dir / JUJU_CONFIG_FILENAME)
    except FileNotFoundError:
        return None
    except OSError as exc:
        raise CraftError(f"Cannot read the {JUJU_CONFIG_FILENAME} file: {exc!r}") from exc

    if allow_broken and (not isinstance(config, dict) or not config.get("options")):
        emit.progress(
            "'config.yaml' is not a valid config file.",
            permanent=True,
        )
        emit.debug(f"Ignoring {JUJU_CONFIG_FILENAME}")
        return None

    emit.debug(f"Validating {JUJU_CONFIG_FILENAME}")
    try:
        return JujuConfig.parse_obj(config)
    except pydantic.ValidationError as error:
        if allow_broken:
            emit.progress(
                format_pydantic_errors(error.errors(), file_name=JUJU_CONFIG_FILENAME),
                permanent=True,
            )
            emit.debug(f"Ignoring {JUJU_CONFIG_FILENAME}")
            return None
        raise


def create_config_yaml(
    basedir: pathlib.Path,
    charmcraft_config: "CharmcraftConfig",
) -> Optional[pathlib.Path]:
    """Create actions.yaml in basedir for given project configuration.

    :param basedir: Directory to create Charm in.
    :param charmcraft_config: Charmcraft configuration object.

    :returns: Path to created config.yaml.
    """
    original_file_path = charmcraft_config.project.dirpath / JUJU_CONFIG_FILENAME
    target_file_path = basedir / JUJU_CONFIG_FILENAME

    # Copy config.yaml if it exists, otherwise create it from CharmcraftConfig.
    if original_file_path.exists():
        # In the build / test process, the original file may be the same as the target file.
        with contextlib.suppress(shutil.SameFileError):
            shutil.copyfile(original_file_path, target_file_path)
    else:
        if charmcraft_config.config:
            target_file_path.write_text(
                yaml.dump(
                    charmcraft_config.config.dict(
                        include={"options"}, exclude_none=True, by_alias=True
                    )
                )
            )
        else:
            return None

    return target_file_path



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Tests for the Store API layer (code in store/store.py)."""

import base64
import platform
from unittest.mock import MagicMock, call, patch

import craft_store
import pytest
from craft_cli import CraftError
from craft_store import attenuations
from craft_store.endpoints import Package
from craft_store.errors import (
    CredentialsAlreadyAvailable,
    CredentialsUnavailable,
    NetworkError,
    StoreServerError,
)
from dateutil import parser

from charmcraft.commands.store.client import AnonymousClient, Client
from charmcraft.commands.store.store import (
    AUTH_DEFAULT_PERMISSIONS,
    AUTH_DEFAULT_TTL,
    Base,
    Library,
    Store,
    _store_client_wrapper,
)
from charmcraft.utils import ResourceOption
from tests.commands.test_store_client import FakeResponse


@pytest.fixture()
def client_mock(monkeypatch):
    """Fixture to provide a mocked client."""
    monkeypatch.setattr(platform, "node", lambda: "fake-host")
    client_mock = MagicMock(spec=Client)
    with patch(
        "charmcraft.commands.store.store.Client", lambda api, storage, ephemeral=True: client_mock
    ):
        yield client_mock


@pytest.fixture()
def anonymous_client_mock(monkeypatch):
    """Fixture to provide a mocked anonymous client."""
    anonymous_client_mock = MagicMock(spec=AnonymousClient)
    with patch(
        "charmcraft.commands.store.store.AnonymousClient",
        lambda api, storage: anonymous_client_mock,
    ):
        yield anonymous_client_mock


# -- tests for client usage


def test_client_init(config):
    """Check that the client is initiated ok even without config."""
    with patch("charmcraft.commands.store.store.Client") as client_mock:
        Store(config.charmhub)
    assert client_mock.mock_calls == [
        call(config.charmhub.api_url, config.charmhub.storage_url, ephemeral=False),
    ]


def test_client_init_ephemeral(config):
    """Check that the client is initiated with no keyring."""
    with patch("charmcraft.commands.store.store.Client") as client_mock:
        Store(config.charmhub, ephemeral=True)
    assert client_mock.mock_calls == [
        call(config.charmhub.api_url, config.charmhub.storage_url, ephemeral=True),
    ]


# -- tests for anonymous client usage


def test_anonymous_client_init(config):
    """Check that the client is initiated ok even without config."""
    with patch("charmcraft.commands.store.store.AnonymousClient") as anonymous_client_mock:
        Store(config.charmhub, needs_auth=False)
    assert anonymous_client_mock.mock_calls == [
        call(config.charmhub.api_url, config.charmhub.storage_url),
    ]


# -- tests for store_client_wrapper


class _FakeAPI:
    def __init__(self, exceptions):
        self.login_called = False
        self.logout_called = False
        self.exceptions = exceptions

    def login(self):
        self.login_called = True

    def logout(self):
        self.logout_called = True

    @_store_client_wrapper()
    def method(self):
        exception = self.exceptions.pop(0)
        if exception is not None:
            raise exception

    @_store_client_wrapper(auto_login=False)
    def method_no_login(self):
        exception = self.exceptions.pop(0)
        if exception is not None:
            raise exception


def test_relogin_on_401_regular_auth(emitter):
    """Tries to re-login after receiving 401 NOT AUTHORIZED from server."""
    api = _FakeAPI([StoreServerError(FakeResponse("auth", 401)), None])

    api.method()

    assert api.login_called is True
    assert api.logout_called is True
    # check logs
    expected = "Existing credentials no longer valid. Trying to log in..."
    emitter.assert_progress(expected)


def test_relogin_on_401_alternate_auth(monkeypatch):
    """Got a 401, but alternate auth is used, so no re-login is done and user is informed."""
    monkeypatch.setenv("CHARMCRAFT_AUTH", "credentials")
    api = _FakeAPI([StoreServerError(FakeResponse("auth", 401)), None])

    with pytest.raises(CraftError) as cm:
        api.method()
    assert str(cm.value) == (
        "Provided credentials are no longer valid for Charmhub. Regenerate them and try again."
    )
    assert api.login_called is False
    assert api.logout_called is False


def test_relogin_on_401_disable_auto_login():
    """Don't try to re-login after receiving 401 NOT AUTHORIZED from server."""
    api = _FakeAPI([StoreServerError(FakeResponse("auth", 401)), None])

    with pytest.raises(CraftError) as cm:
        api.method_no_login()
    assert str(cm.value) == "Existing credentials are no longer valid for Charmhub."
    assert api.login_called is False
    assert api.logout_called is False


def test_relogin_on_401_alternate_auth_disable_auto_login(monkeypatch):
    """Don't try to re-login after receiving 401 NOT AUTHORIZED from server."""
    monkeypatch.setenv("CHARMCRAFT_AUTH", "credentials")
    api = _FakeAPI([StoreServerError(FakeResponse("auth", 401)), None])

    with pytest.raises(CraftError) as cm:
        api.method_no_login()
    assert str(cm.value) == (
        "Provided credentials are no longer valid for Charmhub. Regenerate them and try again."
    )
    assert api.login_called is False
    assert api.logout_called is False


def test_non_401_raises():
    api = _FakeAPI([StoreServerError(FakeResponse("where are you?", 404))])

    with pytest.raises(CraftError) as error:
        api.method()

    assert (
        str(error.value)
        == "Issue encountered while processing your request: [404] where are you?."
    )

    assert api.login_called is False
    assert api.logout_called is False


def test_craft_store_error_raises_command_error():
    api = _FakeAPI([NetworkError(ValueError("network issue"))])

    with pytest.raises(NetworkError) as error:
        api.method()

    assert str(error.value) == "network issue"

    assert api.login_called is False
    assert api.logout_called is False


def test_not_logged_in_warns_regular_auth(emitter):
    """Capture the indication of not being logged in and try to log in."""
    api = _FakeAPI(
        [CredentialsUnavailable(application="charmcraft", host="api.charmhub.io"), None]
    )

    api.method()

    assert api.login_called is True
    assert api.logout_called is False
    # check logs
    expected = "Credentials not found. Trying to log in..."
    emitter.assert_progress(expected)


def test_not_logged_in_warns_alternate_auth(monkeypatch):
    """Capture an indication of not being logged in, which should never happen."""
    monkeypatch.setenv("CHARMCRAFT_AUTH", "credentials")
    api = _FakeAPI(
        [CredentialsUnavailable(application="charmcraft", host="api.charmhub.io"), None]
    )

    with pytest.raises(RuntimeError) as cm:
        api.method()
    assert str(cm.value) == (
        "Charmcraft error: internal inconsistency detected "
        "(CredentialsUnavailable error while having user provided credentials)."
    )
    assert api.login_called is False
    assert api.logout_called is False


def test_not_logged_in_disable_auto_login():
    """Don't try to relogin if not already logged in."""
    api = _FakeAPI(
        [CredentialsUnavailable(application="charmcraft", host="api.charmhub.io"), None]
    )

    with pytest.raises(CredentialsUnavailable):
        api.method_no_login()

    assert api.login_called is False
    assert api.logout_called is False


def test_not_logged_in_alternate_auth_disable_auto_login(monkeypatch):
    """Don't try to relogin if not already logged in."""
    monkeypatch.setenv("CHARMCRAFT_AUTH", "credentials")
    api = _FakeAPI(
        [CredentialsUnavailable(application="charmcraft", host="api.charmhub.io"), None]
    )

    with pytest.raises(RuntimeError) as cm:
        api.method_no_login()
    assert str(cm.value) == (
        "Charmcraft error: internal inconsistency detected "
        "(CredentialsUnavailable error while having user provided credentials)."
    )
    assert api.login_called is False
    assert api.logout_called is False


# -- tests for auth


def test_auth_valid_credentials(config, monkeypatch):
    """No errors raised when initializing Store with valid credentials."""
    monkeypatch.setenv("CHARMCRAFT_AUTH", base64.b64encode(b"good_credentials").decode())
    Store(config.charmhub)


def test_auth_bad_credentials(config, monkeypatch):
    """CraftError raised when initializing Store with bad credentials."""
    monkeypatch.setenv("CHARMCRAFT_AUTH", "bad_credentials")
    with pytest.raises(craft_store.errors.CredentialsNotParseable) as error:
        Store(config.charmhub)

    assert (
        str(error.value) == "Credentials could not be parsed. Expected base64 encoded credentials."
    )


def test_no_keyring(config):
    """Verify CraftStore is raised from Store when no keyring is available."""
    with patch(
        "craft_store.StoreClient.__init__", side_effect=craft_store.errors.NoKeyringError()
    ):
        with pytest.raises(CraftError) as error:
            Store(config.charmhub)

    assert str(error.value) == "No keyring found to store or retrieve credentials from."


def test_login(client_mock, config):
    """Simple login case."""
    # set up a response from client's login
    acquired_credentials = "super secret stuff"
    client_mock.login = MagicMock(return_value=acquired_credentials)

    store = Store(config.charmhub)
    result = store.login()
    assert client_mock.mock_calls == [
        call.login(
            ttl=108000,
            description="charmcraft@fake-host",
            permissions=[
                "account-register-package",
                "account-view-packages",
                "package-manage",
                "package-view",
            ],
        )
    ]
    assert result == acquired_credentials


def test_login_having_credentials(client_mock, config):
    """Login attempt when already having credentials.."""
    # client raises a specific exception for this case
    original_exception = CredentialsAlreadyAvailable("app", "host")
    client_mock.login.side_effect = original_exception

    store = Store(config.charmhub)
    with pytest.raises(CraftError) as cm:
        store.login()
    error = cm.value
    assert str(error) == (
        "Cannot login because credentials were found in your system (which may be "
        "no longer valid, though)."
    )
    assert error.resolution == "Please logout first, then login again."
    assert error.__cause__ is original_exception


def test_login_attenuating_ttl(client_mock, config):
    """Login with specific TTL restrictions."""
    store = Store(config.charmhub)
    store.login(ttl=123)
    assert client_mock.mock_calls == [
        call.login(
            ttl=123,
            description="charmcraft@fake-host",
            permissions=AUTH_DEFAULT_PERMISSIONS,
        )
    ]


def test_login_attenuating_permissions(client_mock, config):
    """Login with specific permissions restrictions."""
    store = Store(config.charmhub)
    permissions_subset = [attenuations.ACCOUNT_VIEW_PACKAGES]
    store.login(permissions=permissions_subset)
    assert client_mock.mock_calls == [
        call.login(
            ttl=AUTH_DEFAULT_TTL,
            description="charmcraft@fake-host",
            permissions=permissions_subset,
        )
    ]


def test_login_attenuating_channels(client_mock, config):
    """Login with specific channels restrictions."""
    store = Store(config.charmhub)
    channels = ["edge", "beta"]
    store.login(channels=channels)
    assert client_mock.mock_calls == [
        call.login(
            ttl=AUTH_DEFAULT_TTL,
            description="charmcraft@fake-host",
            permissions=AUTH_DEFAULT_PERMISSIONS,
            channels=channels,
        )
    ]


def test_login_attenuating_packages(client_mock, config):
    """Login with specific packages restrictions."""
    store = Store(config.charmhub)
    store.login(charms=["supercharm"], bundles=["mybundle1", "mybundle2"])
    assert client_mock.mock_calls == [
        call.login(
            ttl=AUTH_DEFAULT_TTL,
            description="charmcraft@fake-host",
            permissions=AUTH_DEFAULT_PERMISSIONS,
            packages=[
                Package(package_type="charm", package_name="supercharm"),
                Package(package_type="bundle", package_name="mybundle1"),
                Package(package_type="bundle", package_name="mybundle2"),
            ],
        )
    ]


def test_logout(client_mock, config):
    """Simple logout case."""
    store = Store(config.charmhub)
    result = store.logout()
    assert client_mock.mock_calls == [
        call.logout(),
    ]
    assert result is None


def test_whoami_simple(client_mock, config):
    """Simple whoami case."""
    store = Store(config.charmhub)
    auth_response = {
        "account": {
            "display-name": "John Doe",
            "id": "3.14",
            "username": "jdoe",
        },
        "channels": None,
        "packages": None,
        "permissions": ["perm1", "perm2"],
    }
    client_mock.whoami.return_value = auth_response

    result = store.whoami()

    assert client_mock.mock_calls == [
        call.whoami(),
    ]
    assert result.account.name == "John Doe"
    assert result.account.username == "jdoe"
    assert result.account.id == "3.14"
    assert result.channels is None
    assert result.packages is None
    assert result.permissions == ["perm1", "perm2"]


def test_whoami_packages(client_mock, config):
    """Whoami case that specify packages with name or id."""
    store = Store(config.charmhub)
    auth_response = {
        "account": {
            "display-name": "John Doe",
            "id": "3.14",
            "username": "jdoe",
        },
        "channels": None,
        "packages": [
            {"type": "charm", "id": "charmid"},
            {"type": "bundle", "name": "bundlename"},
        ],
        "permissions": ["perm1", "perm2"],
    }
    client_mock.whoami.return_value = auth_response

    result = store.whoami()
    pkg_1, pkg_2 = result.packages
    assert pkg_1.type == "charm"
    assert pkg_1.id == "charmid"
    assert pkg_1.name is None
    assert pkg_2.type == "bundle"
    assert pkg_2.id is None
    assert pkg_2.name == "bundlename"


def test_whoami_channels(client_mock, config):
    """Whoami case with channels indicated."""
    store = Store(config.charmhub)
    auth_response = {
        "account": {
            "display-name": "John Doe",
            "id": "3.14",
            "username": "jdoe",
        },
        "channels": ["edge", "beta"],
        "packages": None,
        "permissions": ["perm1", "perm2"],
    }
    client_mock.whoami.return_value = auth_response

    result = store.whoami()
    assert result.channels == ["edge", "beta"]


# -- tests for register and list names


def test_register_name(client_mock, config):
    """Simple register case."""
    store = Store(config.charmhub)
    result = store.register_name("testname", "stuff")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm", json={"name": "testname", "type": "stuff"}),
    ]
    assert result is None


def test_register_name_unauthorized_logs_in(client_mock, config):
    client_mock.request_urlpath_json.side_effect = [
        StoreServerError(FakeResponse("auth", 401)),
        None,
    ]

    store = Store(config.charmhub)
    store.register_name("testname", "stuff")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm", json={"name": "testname", "type": "stuff"}),
        call.logout(),
        call.login(
            ttl=108000,
            description="charmcraft@fake-host",
            permissions=[
                "account-register-package",
                "account-view-packages",
                "package-manage",
                "package-view",
            ],
        ),
        call.request_urlpath_json("POST", "/v1/charm", json={"name": "testname", "type": "stuff"}),
    ]


# region Unit tests for unregister_name
def test_unregister_name_success(client_mock, config):
    """Simple unregistration."""
    store = Store(config.charmhub)
    store.unregister_name("testname")

    assert client_mock.mock_calls == [call.unregister_name("testname")]


@pytest.mark.parametrize(
    ("http_response", "error_cls"),
    [
        pytest.param(
            FakeResponse("Name testname not found in the charm namespace", 404),
            CraftError,
            id="unknown_name",
        ),
        pytest.param(
            FakeResponse("discharge required", 401), StoreServerError, id="discharge_required"
        ),
        pytest.param(FakeResponse("Unauthorized", 401), StoreServerError, id="Unauthorized"),
        pytest.param(
            FakeResponse("Cannot unregister a package with existing revisions", 403),
            CraftError,
            id="exsting_revision",
        ),
        pytest.param(
            FakeResponse("Charms can only be unregistered by their publisher", 401),
            StoreServerError,
            id="collaborator",
        ),
    ],
)
def test_unregister_name_errors(client_mock, config, http_response: FakeResponse, error_cls):
    """Errors on unregistering a name."""
    client_mock.unregister_name.side_effect = StoreServerError(http_response)

    store = Store(config.charmhub)
    with pytest.raises(error_cls) as exc_info:
        store.unregister_name("testname")

    assert exc_info.value.args[0] == (
        "Issue encountered while processing your request: "
        f"[{http_response.status_code}] {http_response.content}."
    )


@pytest.mark.parametrize(
    "http_response",
    [
        pytest.param(FakeResponse("discharge required", 401), id="discharge_required"),
        pytest.param(FakeResponse("Unauthorized", 401), id="Unauthorized"),
        pytest.param(
            FakeResponse("Charms can only be unregistered by their publisher", 401),
            id="collaborator",
        ),
    ],
)
def test_unregister_name_login(client_mock, config, http_response: FakeResponse):
    """Retry login when registering a name."""
    client_mock.unregister_name.side_effect = [StoreServerError(http_response), None]

    store = Store(config.charmhub)
    store.unregister_name("testname")


# endregion


def test_list_registered_names_empty(client_mock, config):
    """List registered names getting an empty response."""
    store = Store(config.charmhub)

    auth_response = {"results": []}
    client_mock.request_urlpath_json.return_value = auth_response

    result = store.list_registered_names(include_collaborations=False)

    assert client_mock.mock_calls == [call.request_urlpath_json("GET", "/v1/charm")]
    assert result == []


def test_list_registered_names_multiple(client_mock, config):
    """List registered names getting a multiple response."""
    store = Store(config.charmhub)

    publisher = {"display-name": "J. Doe", "other-info": "a lot"}
    auth_response = {
        "results": [
            {
                "name": "name1",
                "type": "charm",
                "private": False,
                "status": "status1",
                "publisher": publisher,
            },
            {
                "name": "name2",
                "type": "bundle",
                "private": True,
                "status": "status2",
                "publisher": publisher,
            },
        ]
    }
    client_mock.request_urlpath_json.return_value = auth_response

    result = store.list_registered_names(include_collaborations=False)

    assert client_mock.mock_calls == [call.request_urlpath_json("GET", "/v1/charm")]
    item1, item2 = result
    assert item1.name == "name1"
    assert item1.entity_type == "charm"
    assert not item1.private
    assert item1.status == "status1"
    assert item1.publisher_display_name == "J. Doe"
    assert item2.name == "name2"
    assert item2.entity_type == "bundle"
    assert item2.private
    assert item2.status == "status2"
    assert item2.publisher_display_name == "J. Doe"


def test_list_registered_names_include_collaborations(client_mock, config):
    """List registered names including collaborations."""
    store = Store(config.charmhub)

    auth_response = {
        "results": [
            {
                "name": "name1",
                "type": "charm",
                "private": False,
                "status": "status1",
                "publisher": {"display-name": "J. Doe", "other-info": "a lot"},
            },
            {
                "name": "name2",
                "type": "bundle",
                "private": True,
                "status": "status2",
                "publisher": {"display-name": "Anonymous", "other-info": "more"},
            },
        ]
    }
    client_mock.request_urlpath_json.return_value = auth_response

    result = store.list_registered_names(include_collaborations=True)

    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm?include-collaborations=true")
    ]
    item1, item2 = result
    assert item1.name == "name1"
    assert item1.entity_type == "charm"
    assert not item1.private
    assert item1.status == "status1"
    assert item1.publisher_display_name == "J. Doe"
    assert item2.name == "name2"
    assert item2.entity_type == "bundle"
    assert item2.private
    assert item2.status == "status2"
    assert item2.publisher_display_name == "Anonymous"


# -- tests for the upload functionality (both for charm/bundles and resources)


def test_upload_straightforward(client_mock, emitter, config):
    """The full and successful upload case."""
    store = Store(config.charmhub)

    # the first response, for when pushing bytes
    test_upload_id = "test-upload-id"
    client_mock.push_file.return_value = test_upload_id

    # the second response, for telling the store it was pushed
    test_status_url = "https://store.c.c/status"

    # the third response, status ok (note the patched UPLOAD_ENDING_STATUSES below)
    test_revision = 123
    test_status_ok = "test-status"
    status_response = {
        "revisions": [{"status": test_status_ok, "revision": test_revision, "errors": None}]
    }

    client_mock.request_urlpath_json.side_effect = [
        {"status-url": test_status_url},
        status_response,
    ]

    test_status_resolution = "test-ok-or-not"
    fake_statuses = {test_status_ok: test_status_resolution}
    test_filepath = "test-filepath"
    test_endpoint = "/v1/test/revisions/endpoint/"
    with patch.dict("charmcraft.commands.store.store.UPLOAD_ENDING_STATUSES", fake_statuses):
        result = store._upload(test_endpoint, test_filepath)

    # check all client calls
    assert client_mock.mock_calls == [
        call.whoami(),
        call.push_file(test_filepath),
        call.request_urlpath_json("POST", test_endpoint, json={"upload-id": test_upload_id}),
        call.request_urlpath_json("GET", test_status_url),
    ]

    # check result (build after patched ending struct)
    assert result.ok == test_status_resolution
    assert result.status == test_status_ok
    assert result.revision == test_revision

    # check logs
    emitter.assert_interactions(
        [
            call(
                "progress",
                "Upload test-upload-id started, got status url https://store.c.c/status",
            ),
            call("progress", "Status checked: " + str(status_response)),
        ]
    )


def test_upload_polls_status_ok(client_mock, emitter, config):
    """Upload polls status url until the end is indicated."""
    store = Store(config.charmhub)

    # first and second response, for pushing bytes and let the store know about it
    test_upload_id = "test-upload-id"
    client_mock.push_file.return_value = test_upload_id
    test_status_url = "https://store.c.c/status"

    # the status checking response, will answer something not done yet twice, then ok
    test_revision = 123
    test_status_ok = "test-status"
    status_response_1 = {
        "revisions": [{"status": "still-scanning", "revision": None, "errors": None}]
    }
    status_response_2 = {
        "revisions": [{"status": "more-revisions", "revision": None, "errors": None}]
    }
    status_response_3 = {
        "revisions": [{"status": test_status_ok, "revision": test_revision, "errors": None}]
    }
    client_mock.request_urlpath_json.side_effect = [
        {"status-url": test_status_url},
        status_response_1,
        status_response_2,
        status_response_3,
    ]

    test_status_resolution = "clean and crispy"
    fake_statuses = {test_status_ok: test_status_resolution}
    with patch.dict("charmcraft.commands.store.store.UPLOAD_ENDING_STATUSES", fake_statuses):
        with patch("charmcraft.commands.store.store.POLL_DELAYS", [0.1] * 5):
            result = store._upload("/test/endpoint/", "some-filepath")

    # check the status-checking client calls (kept going until third one)
    assert client_mock.mock_calls[3:] == [
        call.request_urlpath_json("GET", test_status_url),
        call.request_urlpath_json("GET", test_status_url),
        call.request_urlpath_json("GET", test_status_url),
    ]

    # check result which must have values from final result
    assert result.ok == test_status_resolution
    assert result.status == test_status_ok
    assert result.revision == test_revision

    # check logs
    emitter.assert_interactions(
        [
            call(
                "progress",
                "Upload test-upload-id started, got status url https://store.c.c/status",
            ),
            call("progress", "Status checked: " + str(status_response_1)),
            call("progress", "Status checked: " + str(status_response_2)),
            call("progress", "Status checked: " + str(status_response_3)),
        ]
    )


def test_upload_polls_status_timeout(client_mock, emitter, config):
    """Upload polls status url until a timeout is achieved.

    This is simulated patching a POLL_DELAYS structure shorter than the
    number of "keep going" responses.
    """
    store = Store(config.charmhub)

    # first and second response, for pushing bytes and let the store know about it
    test_upload_id = "test-upload-id"
    client_mock.push_file.return_value = test_upload_id
    test_status_url = "https://store.c.c/status"

    # the status checking response, will answer something not done yet twice, then ok
    test_status_ok = "test-status"
    status_response = {
        "revisions": [{"status": "still-scanning", "revision": None, "errors": None}]
    }
    client_mock.request_urlpath_json.side_effect = [
        {"status-url": test_status_url},
        status_response,
        status_response,
        status_response,
    ]

    test_status_resolution = "clean and crispy"
    fake_statuses = {test_status_ok: test_status_resolution}
    with patch.dict("charmcraft.commands.store.store.UPLOAD_ENDING_STATUSES", fake_statuses):
        with patch("charmcraft.commands.store.store.POLL_DELAYS", [0.1] * 2):
            with pytest.raises(CraftError) as cm:
                store._upload("/test/endpoint/", "some-filepath")
    assert str(cm.value) == "Timeout polling Charmhub for upload status (after 0.2s)."


def test_upload_error(client_mock, config):
    """The upload ended in error."""
    store = Store(config.charmhub)

    # the first response, for when pushing bytes
    test_upload_id = "test-upload-id"
    client_mock.push_file.return_value = test_upload_id

    # the second response, for telling the store it was pushed
    test_status_url = "https://store.c.c/status"

    # the third response, status in error (note the patched UPLOAD_ENDING_STATUSES below)
    test_revision = 123
    test_status_bad = "test-status"
    status_response = {
        "revisions": [
            {
                "status": test_status_bad,
                "revision": test_revision,
                "errors": [
                    {"message": "error text 1", "code": "error-code-1"},
                    {"message": "error text 2", "code": "error-code-2"},
                ],
            }
        ]
    }

    client_mock.request_urlpath_json.side_effect = [
        {"status-url": test_status_url},
        status_response,
    ]

    test_status_resolution = "test-ok-or-not"
    fake_statuses = {test_status_bad: test_status_resolution}
    test_filepath = "test-filepath"
    with patch.dict("charmcraft.commands.store.store.UPLOAD_ENDING_STATUSES", fake_statuses):
        result = store._upload("/test/endpoint/", test_filepath)

    # check result
    assert result.ok == test_status_resolution
    assert result.status == test_status_bad
    assert result.revision == test_revision
    error1, error2 = result.errors
    assert error1.message == "error text 1"
    assert error1.code == "error-code-1"
    assert error2.message == "error text 2"
    assert error2.code == "error-code-2"


@pytest.mark.usefixtures("client_mock")
def test_upload_charmbundles_endpoint(config):
    """The bundle/charm upload prepares ok the endpoint and calls the generic _upload."""
    store = Store(config.charmhub)
    test_results = "test-results"

    with patch.object(store, "_upload") as mock:
        mock.return_value = test_results
        result = store.upload("test-charm", "test-filepath")
    mock.assert_called_once_with("/v1/charm/test-charm/revisions", "test-filepath")
    assert result == test_results


@pytest.mark.usefixtures("client_mock")
def test_upload_resources_endpoint(config):
    """The resource upload prepares ok the endpoint and calls the generic _upload."""
    store = Store(config.charmhub)
    test_results = "test-results"

    with patch.object(store, "_upload") as mock:
        mock.return_value = test_results
        result = store.upload_resource("test-charm", "test-resource", "test-type", "test-filepath")
    expected_endpoint = "/v1/charm/test-charm/resources/test-resource/revisions"
    mock.assert_called_once_with(
        expected_endpoint, "test-filepath", extra_fields={"type": "test-type"}
    )
    assert result == test_results


def test_upload_including_extra_parameters(client_mock, emitter, config):
    """Verify that the upload includes extra parameters if given."""
    store = Store(config.charmhub)

    # the first response, for when pushing bytes
    test_upload_id = "test-upload-id"
    client_mock.push_file.return_value = test_upload_id

    # the second response, for telling the store it was pushed
    test_status_url = "https://store.c.c/status"

    # the third response, status ok (note the patched UPLOAD_ENDING_STATUSES below)
    test_revision = 123
    test_status_ok = "test-status"
    status_response = {
        "revisions": [{"status": test_status_ok, "revision": test_revision, "errors": None}]
    }

    client_mock.request_urlpath_json.side_effect = [
        {"status-url": test_status_url},
        status_response,
    ]

    test_status_resolution = "test-ok-or-not"
    fake_statuses = {test_status_ok: test_status_resolution}
    test_filepath = "test-filepath"
    test_endpoint = "/v1/test/revisions/endpoint/"
    extra_fields = {"extra-key": "1", "more": "2"}
    with patch.dict("charmcraft.commands.store.store.UPLOAD_ENDING_STATUSES", fake_statuses):
        store._upload(test_endpoint, test_filepath, extra_fields=extra_fields)

    # check all client calls
    assert client_mock.mock_calls == [
        call.whoami(),
        call.push_file(test_filepath),
        call.request_urlpath_json(
            "POST",
            test_endpoint,
            json={"upload-id": test_upload_id, "extra-key": "1", "more": "2"},
        ),
        call.request_urlpath_json("GET", test_status_url),
    ]


# -- tests for list revisions


def test_list_revisions_ok(client_mock, config):
    """One revision ok."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {
        "revisions": [
            {
                "revision": 7,
                "version": "v7",
                "created-at": "2020-06-29T22:11:00.123",
                "status": "approved",
                "errors": None,
                "bases": [{"architecture": "amd64", "channel": "20.04", "name": "ubuntu"}],
            }
        ]
    }

    result = store.list_revisions("some-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm/some-name/revisions")
    ]

    (item,) = result
    assert item.revision == 7
    assert item.version == "v7"
    assert item.created_at == parser.parse("2020-06-29T22:11:00.123")
    assert item.status == "approved"
    assert item.errors == []
    assert item.bases == [Base(architecture="amd64", channel="20.04", name="ubuntu")]


def test_list_revisions_empty(client_mock, config):
    """No revisions listed."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {"revisions": []}

    result = store.list_revisions("some-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm/some-name/revisions")
    ]
    assert result == []


def test_list_revisions_errors(client_mock, config):
    """One revision with errors."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {
        "revisions": [
            {
                "revision": 7,
                "version": "v7",
                "created-at": "2020-06-29T22:11:00.123",
                "status": "rejected",
                "errors": [
                    {"message": "error text 1", "code": "error-code-1"},
                    {"message": "error text 2", "code": "error-code-2"},
                ],
                "bases": [{"architecture": "amd64", "channel": "20.04", "name": "ubuntu"}],
            }
        ]
    }

    result = store.list_revisions("some-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm/some-name/revisions")
    ]

    (item,) = result
    error1, error2 = item.errors
    assert error1.message == "error text 1"
    assert error1.code == "error-code-1"
    assert error2.message == "error text 2"
    assert error2.code == "error-code-2"


def test_list_revisions_several_mixed(client_mock, config):
    """All cases mixed."""
    client_mock.request_urlpath_json.return_value = {
        "revisions": [
            {
                "revision": 1,
                "version": "v1",
                "created-at": "2020-06-29T22:11:01",
                "status": "rejected",
                "errors": [
                    {"message": "error", "code": "code"},
                ],
                "bases": [{"architecture": "amd64", "channel": "20.04", "name": "ubuntu"}],
            },
            {
                "revision": 2,
                "version": "v2",
                "created-at": "2020-06-29T22:11:02",
                "status": "approved",
                "errors": None,
                "bases": [{"architecture": "amd64", "channel": "20.04", "name": "ubuntu"}],
            },
        ]
    }

    store = Store(config.charmhub)
    result = store.list_revisions("some-name")

    (item1, item2) = result

    assert item1.revision == 1
    assert item1.version == "v1"
    assert item1.created_at == parser.parse("2020-06-29T22:11:01")
    assert item1.status == "rejected"
    (error,) = item1.errors
    assert error.message == "error"
    assert error.code == "code"

    assert item2.revision == 2
    assert item2.version == "v2"
    assert item2.created_at == parser.parse("2020-06-29T22:11:02")
    assert item2.status == "approved"
    assert item2.errors == []


def test_list_revisions_bases_none(client_mock, config):
    """Bases in None answered by the store (happens with bundles, for example)."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {
        "revisions": [
            {
                "revision": 7,
                "version": "v7",
                "created-at": "2020-06-29T22:11:00.123",
                "status": "approved",
                "errors": None,
                "bases": [None],
            }
        ]
    }
    result = store.list_revisions("some-name")
    (item,) = result
    assert item.bases == [None]


# -- tests for release


def test_release_simple(client_mock, config):
    """Releasing a revision into one channel."""
    store = Store(config.charmhub)
    store.release("testname", 123, ["somechannel"], [])

    expected_body = [{"revision": 123, "channel": "somechannel", "resources": []}]
    assert client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm/testname/releases", json=expected_body),
    ]


def test_release_multiple_channels(client_mock, config):
    """Releasing a revision into multiple channels."""
    store = Store(config.charmhub)
    store.release("testname", 123, ["channel1", "channel2", "channel3"], [])

    expected_body = [
        {"revision": 123, "channel": "channel1", "resources": []},
        {"revision": 123, "channel": "channel2", "resources": []},
        {"revision": 123, "channel": "channel3", "resources": []},
    ]
    assert client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm/testname/releases", json=expected_body),
    ]


def test_release_with_resources(client_mock, config):
    """Releasing with resources attached."""
    store = Store(config.charmhub)
    r1 = ResourceOption(name="foo", revision=3)
    r2 = ResourceOption(name="bar", revision=17)
    store.release("testname", 123, ["channel1", "channel2"], [r1, r2])

    expected_body = [
        {
            "revision": 123,
            "channel": "channel1",
            "resources": [
                {"name": "foo", "revision": 3},
                {"name": "bar", "revision": 17},
            ],
        },
        {
            "revision": 123,
            "channel": "channel2",
            "resources": [
                {"name": "foo", "revision": 3},
                {"name": "bar", "revision": 17},
            ],
        },
    ]
    assert client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm/testname/releases", json=expected_body),
    ]


# -- tests for status


def test_status_ok(client_mock, config):
    """Get all the release information."""
    client_mock.request_urlpath_json.return_value = {
        "channel-map": [
            {
                "channel": "latest/beta",
                "expiration-date": None,
                "progressive": {"paused": None, "percentage": None},
                "revision": 5,
                "when": "2020-07-16T18:45:24Z",
                "resources": [],
                "base": {"architecture": "amd64", "channel": "20.04", "name": "ubuntu"},
            },
            {
                "channel": "latest/edge/mybranch",
                "expiration-date": "2020-08-16T18:46:02Z",
                "progressive": {"paused": None, "percentage": None},
                "revision": 10,
                "when": "2020-07-16T18:46:02Z",
                "resources": [],
                "base": {"architecture": "amd64", "channel": "20.04", "name": "ubuntu"},
            },
        ],
        "package": {
            "channels": [
                {
                    "branch": None,
                    "fallback": None,
                    "name": "latest/stable",
                    "risk": "stable",
                    "track": "latest",
                },
                {
                    "branch": "mybranch",
                    "fallback": "latest/stable",
                    "name": "latest/edge/mybranch",
                    "risk": "edge",
                    "track": "latest",
                },
            ]
        },
        "revisions": [
            {
                "revision": 5,
                "version": "5",
                "created-at": "2020-06-29T22:11:05",
                "status": "approved",
                "errors": None,
                "bases": [{"architecture": "amd64", "channel": "20.04", "name": "ubuntu"}],
            },
            {
                "revision": 10,
                "version": "63a852b",
                "created-at": "2020-06-29T22:11:10",
                "status": "approved",
                "errors": None,
                "bases": [{"architecture": "amd64", "channel": "20.04", "name": "ubuntu"}],
            },
        ],
    }

    store = Store(config.charmhub)
    channel_map, channels, revisions = store.list_releases("testname")

    # check how the client is used
    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm/testname/releases"),
    ]

    # check response
    cmap1, cmap2 = channel_map
    assert cmap1.revision == 5
    assert cmap1.channel == "latest/beta"
    assert cmap1.expires_at is None
    assert cmap1.resources == []
    assert cmap1.base.name == "ubuntu"
    assert cmap1.base.channel == "20.04"
    assert cmap1.base.architecture == "amd64"
    assert cmap2.revision == 10
    assert cmap2.channel == "latest/edge/mybranch"
    assert cmap2.expires_at == parser.parse("2020-08-16T18:46:02Z")
    assert cmap2.resources == []
    assert cmap2.base.name == "ubuntu"
    assert cmap2.base.channel == "20.04"
    assert cmap2.base.architecture == "amd64"

    channel1, channel2 = channels
    assert channel1.name == "latest/stable"
    assert channel1.track == "latest"
    assert channel1.risk == "stable"
    assert channel1.branch is None
    assert channel2.name == "latest/edge/mybranch"
    assert channel2.track == "latest"
    assert channel2.risk == "edge"
    assert channel2.branch == "mybranch"

    rev1, rev2 = revisions
    assert rev1.revision == 5
    assert rev1.version == "5"
    assert rev1.created_at == parser.parse("2020-06-29T22:11:05")
    assert rev1.status == "approved"
    assert rev1.errors == []
    (base,) = rev1.bases
    assert base.name == "ubuntu"
    assert base.channel == "20.04"
    assert base.architecture == "amd64"
    assert rev2.revision == 10
    assert rev2.version == "63a852b"
    assert rev2.created_at == parser.parse("2020-06-29T22:11:10")
    assert rev2.status == "approved"
    assert rev2.errors == []
    (base,) = rev2.bases
    assert base.name == "ubuntu"
    assert base.channel == "20.04"
    assert base.architecture == "amd64"


def test_status_with_resources(client_mock, config):
    """Get all the release information."""
    client_mock.request_urlpath_json.return_value = {
        "channel-map": [
            {
                "channel": "latest/stable",
                "expiration-date": None,
                "progressive": {"paused": None, "percentage": None},
                "revision": 5,
                "when": "2020-07-16T18:45:24Z",
                "resources": [
                    {
                        "name": "test-resource-1",
                        "revision": 2,
                        "type": "file",
                    },
                ],
                "base": {"architecture": "amd64", "channel": "20.04", "name": "ubuntu"},
            },
            {
                "channel": "latest/edge",
                "expiration-date": "2020-08-16T18:46:02Z",
                "progressive": {"paused": None, "percentage": None},
                "revision": 5,
                "when": "2020-07-16T18:46:02Z",
                "resources": [
                    {
                        "name": "test-resource-1",
                        "revision": 2,
                        "type": "file",
                    },
                    {
                        "name": "test-resource-2",
                        "revision": 329,
                        "type": "file",
                    },
                ],
                "base": {"architecture": "amd64", "channel": "20.04", "name": "ubuntu"},
            },
        ],
        "package": {
            "channels": [
                {
                    "branch": None,
                    "fallback": None,
                    "name": "latest/edge",
                    "risk": "edge",
                    "track": "latest",
                },
                {
                    "branch": None,
                    "fallback": None,
                    "name": "latest/stable",
                    "risk": "stable",
                    "track": "latest",
                },
            ]
        },
        "revisions": [
            {
                "revision": 5,
                "version": "5",
                "created-at": "2020-06-29T22:11:05",
                "status": "approved",
                "errors": None,
                "bases": [{"architecture": "amd64", "channel": "20.04", "name": "ubuntu"}],
            },
        ],
    }

    store = Store(config.charmhub)
    channel_map, _, _ = store.list_releases("testname")

    # check response
    cmap1, cmap2 = channel_map

    assert cmap1.revision == 5
    assert cmap1.channel == "latest/stable"
    assert cmap1.expires_at is None
    (res,) = cmap1.resources
    assert res.name == "test-resource-1"
    assert res.revision == 2
    assert res.resource_type == "file"

    assert cmap2.revision == 5
    assert cmap2.channel == "latest/edge"
    assert cmap2.expires_at == parser.parse("2020-08-16T18:46:02Z")
    (res1, res2) = cmap2.resources
    assert res1.name == "test-resource-1"
    assert res1.revision == 2
    assert res1.resource_type == "file"
    assert res2.name == "test-resource-2"
    assert res2.revision == 329
    assert res2.resource_type == "file"


def test_status_base_in_None(client_mock, config):
    """Support the case of base being None (may happen with bundles)."""
    client_mock.request_urlpath_json.return_value = {
        "channel-map": [
            {
                "channel": "latest/stable",
                "expiration-date": None,
                "progressive": {"paused": None, "percentage": None},
                "revision": 5,
                "when": "2020-07-16T18:45:24Z",
                "resources": [],
                "base": None,
            },
        ],
        "package": {
            "channels": [
                {
                    "branch": None,
                    "fallback": None,
                    "name": "latest/stable",
                    "risk": "stable",
                    "track": "latest",
                },
            ]
        },
        "revisions": [
            {
                "revision": 5,
                "version": "5",
                "created-at": "2020-06-29T22:11:05",
                "status": "approved",
                "errors": None,
                "bases": [None],
            },
        ],
    }

    store = Store(config.charmhub)
    channel_map, _, revisions = store.list_releases("testname")

    # check response
    (cmap,) = channel_map
    assert cmap.base is None
    (rev,) = revisions
    rev.bases == [None]


# -- tests for library related functions


def test_create_library_id(client_mock, config):
    """Create a new library in the store."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {"library-id": "test-lib-id"}

    result = store.create_library_id("test-charm-name", "test-lib-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json(
            "POST",
            "/v1/charm/libraries/test-charm-name",
            json={"library-name": "test-lib-name"},
        ),
    ]
    assert result == "test-lib-id"


def test_create_library_revision(client_mock, config):
    """Create a new library revision in the store."""
    test_charm_name = "test-charm-name"
    test_lib_name = "test-lib-name"
    test_lib_id = "test-lib-id"
    test_api = "test-api-version"
    test_patch = "test-patch-version"
    test_content = "test content with quite a lot of funny Python code :p"
    test_hash = "1234"

    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {
        "api": test_api,
        "content": test_content,
        "hash": test_hash,
        "library-id": test_lib_id,
        "library-name": test_lib_name,
        "charm-name": test_charm_name,
        "patch": test_patch,
    }

    result_lib = store.create_library_revision(
        test_charm_name, test_lib_id, test_api, test_patch, test_content, test_hash
    )

    payload = {
        "api": test_api,
        "patch": test_patch,
        "content": test_content,
        "hash": test_hash,
    }
    assert client_mock.mock_calls == [
        call.request_urlpath_json(
            "POST", "/v1/charm/libraries/test-charm-name/" + test_lib_id, json=payload
        ),
    ]
    assert result_lib.api == test_api
    assert result_lib.content == test_content
    assert result_lib.content_hash == test_hash
    assert result_lib.lib_id == test_lib_id
    assert result_lib.lib_name == test_lib_name
    assert result_lib.charm_name == test_charm_name
    assert result_lib.patch == test_patch


def test_get_library(anonymous_client_mock, config):
    """Get all the information (including content) for a library revision."""
    test_charm_name = "test-charm-name"
    test_lib_name = "test-lib-name"
    test_lib_id = "test-lib-id"
    test_api = "test-api-version"
    test_patch = "test-patch-version"
    test_content = "test content with quite a lot of funny Python code :p"
    test_hash = "1234"

    store = Store(config.charmhub, needs_auth=False)
    anonymous_client_mock.request_urlpath_json.return_value = {
        "api": test_api,
        "content": test_content,
        "hash": test_hash,
        "library-id": test_lib_id,
        "library-name": test_lib_name,
        "charm-name": test_charm_name,
        "patch": test_patch,
    }

    result_lib = store.get_library(test_charm_name, test_lib_id, test_api)

    assert anonymous_client_mock.mock_calls == [
        call.request_urlpath_json(
            "GET", f"/v1/charm/libraries/test-charm-name/{test_lib_id}?api={test_api}"
        ),
    ]
    assert result_lib.api == test_api
    assert result_lib.content == test_content
    assert result_lib.content_hash == test_hash
    assert result_lib.lib_id == test_lib_id
    assert result_lib.lib_name == test_lib_name
    assert result_lib.charm_name == test_charm_name
    assert result_lib.patch == test_patch


def test_get_tips_simple(anonymous_client_mock, config):
    """Get info for a lib, simple case with successful result."""
    test_charm_name = "test-charm-name"
    test_lib_name = "test-lib-name"
    test_lib_id = "test-lib-id"
    test_api = "test-api-version"
    test_patch = "test-patch-version"
    test_content = "test content with quite a lot of funny Python code :p"
    test_hash = "1234"

    store = Store(config.charmhub, needs_auth=False)
    anonymous_client_mock.request_urlpath_json.return_value = {
        "libraries": [
            {
                "api": test_api,
                "content": test_content,
                "hash": test_hash,
                "library-id": test_lib_id,
                "library-name": test_lib_name,
                "charm-name": test_charm_name,
                "patch": test_patch,
            }
        ]
    }

    query_info = [
        {"lib_id": test_lib_id},
    ]
    result = store.get_libraries_tips(query_info)

    payload = [
        {"library-id": test_lib_id},
    ]
    assert anonymous_client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm/libraries/bulk", json=payload),
    ]
    expected = {
        (test_lib_id, test_api): Library(
            api=test_api,
            content=test_content,
            content_hash=test_hash,
            lib_id=test_lib_id,
            lib_name=test_lib_name,
            charm_name=test_charm_name,
            patch=test_patch,
        ),
    }
    assert result == expected


def test_get_tips_empty(anonymous_client_mock, config):
    """Get info for a lib, with an empty response."""
    test_lib_id = "test-lib-id"

    store = Store(config.charmhub, needs_auth=False)
    anonymous_client_mock.request_urlpath_json.return_value = {"libraries": []}

    query_info = [
        {"lib_id": test_lib_id},
    ]
    result = store.get_libraries_tips(query_info)

    payload = [
        {"library-id": test_lib_id},
    ]
    assert anonymous_client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm/libraries/bulk", json=payload),
    ]
    assert result == {}


def test_get_tips_several(anonymous_client_mock, config):
    """Get info for multiple libs at once."""
    test_charm_name_1 = "test-charm-name-1"
    test_lib_name_1 = "test-lib-name-1"
    test_lib_id_1 = "test-lib-id-1"
    test_api_1 = "test-api-version-1"
    test_patch_1 = "test-patch-version-1"
    test_content_1 = "test content with quite a lot of funny Python code :p"
    test_hash_1 = "1234"

    test_charm_name_2 = "test-charm-name-2"
    test_lib_name_2 = "test-lib-name-2"
    test_lib_id_2 = "test-lib-id-2"
    test_api_2 = "test-api-version-2"
    test_patch_2 = "test-patch-version-2"
    test_content_2 = "more awesome Python code :)"
    test_hash_2 = "5678"

    store = Store(config.charmhub, needs_auth=False)
    anonymous_client_mock.request_urlpath_json.return_value = {
        "libraries": [
            {
                "api": test_api_1,
                "content": test_content_1,
                "hash": test_hash_1,
                "library-id": test_lib_id_1,
                "library-name": test_lib_name_1,
                "charm-name": test_charm_name_1,
                "patch": test_patch_1,
            },
            {
                "api": test_api_2,
                "content": test_content_2,
                "hash": test_hash_2,
                "library-id": test_lib_id_2,
                "library-name": test_lib_name_2,
                "charm-name": test_charm_name_2,
                "patch": test_patch_2,
            },
        ]
    }

    query_info = [
        {"lib_id": test_lib_id_1},
        {"lib_id": test_lib_id_2},
    ]
    result = store.get_libraries_tips(query_info)

    payload = [
        {"library-id": test_lib_id_1},
        {"library-id": test_lib_id_2},
    ]
    assert anonymous_client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm/libraries/bulk", json=payload),
    ]
    expected = {
        (test_lib_id_1, test_api_1): Library(
            api=test_api_1,
            content=test_content_1,
            content_hash=test_hash_1,
            lib_id=test_lib_id_1,
            lib_name=test_lib_name_1,
            charm_name=test_charm_name_1,
            patch=test_patch_1,
        ),
        (test_lib_id_2, test_api_2): Library(
            api=test_api_2,
            content=test_content_2,
            content_hash=test_hash_2,
            lib_id=test_lib_id_2,
            lib_name=test_lib_name_2,
            charm_name=test_charm_name_2,
            patch=test_patch_2,
        ),
    }
    assert result == expected


def test_get_tips_query_combinations(anonymous_client_mock, config):
    """Use all the combinations to specify what's queried."""
    store = Store(config.charmhub, needs_auth=False)
    anonymous_client_mock.request_urlpath_json.return_value = {"libraries": []}

    query_info = [
        {"lib_id": "test-lib-id-1"},
        {"lib_id": "test-lib-id-2", "api": 2},
        {"charm_name": "test-charm-name-3"},
        {"charm_name": "test-charm-name-4", "api": 4},
        {"charm_name": "test-charm-name-5", "lib_name": "test-lib-name-5"},
        {"charm_name": "test-charm-name-6", "lib_name": "test-lib-name-6", "api": 6},
    ]
    store.get_libraries_tips(query_info)

    payload = [
        {"library-id": "test-lib-id-1"},
        {"library-id": "test-lib-id-2", "api": 2},
        {"charm-name": "test-charm-name-3"},
        {"charm-name": "test-charm-name-4", "api": 4},
        {"charm-name": "test-charm-name-5", "library-name": "test-lib-name-5"},
        {
            "charm-name": "test-charm-name-6",
            "library-name": "test-lib-name-6",
            "api": 6,
        },
    ]
    assert anonymous_client_mock.mock_calls == [
        call.request_urlpath_json("POST", "/v1/charm/libraries/bulk", json=payload),
    ]


# -- tests for list resources


def test_list_resources_ok(client_mock, config):
    """One resource ok."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {
        "resources": [
            {
                "name": "testresource",
                "optional": True,
                "revision": 9,
                "type": "file",
            },
        ]
    }

    result = store.list_resources("some-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm/some-name/resources")
    ]

    (item,) = result
    assert item.name == "testresource"
    assert item.optional
    assert item.revision == 9
    assert item.resource_type == "file"


def test_list_resources_empty(client_mock, config):
    """No resources listed."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {"resources": []}

    result = store.list_resources("some-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm/some-name/resources")
    ]
    assert result == []


def test_list_resources_several(client_mock, config):
    """Several items returned."""
    client_mock.request_urlpath_json.return_value = {
        "resources": [
            {
                "name": "testresource1",
                "optional": True,
                "revision": 123,
                "type": "file",
            },
            {
                "name": "testresource2",
                "optional": False,
                "revision": 678,
                "type": "file",
            },
        ]
    }

    store = Store(config.charmhub)
    result = store.list_resources("some-name")

    (item1, item2) = result

    assert item1.name == "testresource1"
    assert item1.optional is True
    assert item1.revision == 123
    assert item1.resource_type == "file"

    assert item2.name == "testresource2"
    assert item2.optional is False
    assert item2.revision == 678
    assert item2.resource_type == "file"


# -- tests for list resource revisions


def test_list_resource_revisions_ok(client_mock, config):
    """One resource revision ok."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {
        "revisions": [
            {
                "created-at": "2021-02-11T13:43:22.396606",
                "name": "otherstuff",
                "revision": 1,
                "sha256": "1bf0399c2de1240777ba73785f1ff1de5331f12853765a0",
                "sha3-384": "deb9369cb2b9e86ad44160e93da43d240e6388c5dc67b8e2a5a3c2a36a26fe4c89",
                "sha384": "eaaba6aa119da415e6ad778358a8530c47fefbe3ceced258e8c25530107dc7908e",
                "sha512": (
                    "b8cfe885d49285d8546885167a72fd56ea23480e17c9cdd8e06b45239d79b774c6d6fc09d"
                ),
                "size": 500,
            },
        ]
    }

    result = store.list_resource_revisions("charm-name", "resource-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm/charm-name/resources/resource-name/revisions")
    ]

    (item,) = result
    assert item.revision == 1
    assert item.created_at == parser.parse("2021-02-11T13:43:22.396606")
    assert item.size == 500


def test_list_resource_revisions_empty(client_mock, config):
    """No resource revisions listed."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {"revisions": []}

    result = store.list_resource_revisions("charm-name", "resource-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json("GET", "/v1/charm/charm-name/resources/resource-name/revisions")
    ]
    assert result == []


def test_list_resource_revisions_several(client_mock, config):
    """Several items returned."""
    client_mock.request_urlpath_json.return_value = {
        "revisions": [
            {
                "created-at": "2021-02-11T13:43:22.396606",
                "name": "otherstuff",
                "revision": 1,
                "sha256": "1bf0399c2de1240777ba73785f1ff1de5331f12853765a0",
                "sha3-384": "deb9369cb2b9e86ad44160e93da43d240e6388c5dc67b8e2a5a3c2a36a26fe4c89",
                "sha384": "eaaba6aa119da415e6ad778358a8530c47fefbe3ceced258e8c25530107dc7908e",
                "sha512": (
                    "b8cfe885d49285d8546885167a72fd56ea23480e17c9cdd8e06b45239d79b774c6d6fc09d"
                ),
                "size": 500,
            },
            {
                "created-at": "2021-02-11T14:23:55.659148",
                "name": "otherstuff",
                "revision": 2,
                "sha256": "73785f1ff1de5331f12853765a01bf0399c2de1240777ba",
                "sha3-384": "60e93da43d240e6388c5dc67b8e2a5a3c2a36a26fe4c89deb9369cb2b5e86ad441",
                "sha384": "778358a8530c47fefbe3ceced258e8c25530107dc7908eeaaba6aa119dad15e6ad",
                "sha512": (
                    "05167a72fd56ea23480e17c9cdd8e06b45239d79b774c6d6fc09db8cfe885d49285d8547c"
                ),
                "size": 420,
            },
        ]
    }

    store = Store(config.charmhub)
    result = store.list_resource_revisions("charm-name", "resource-name")

    (item1, item2) = result

    assert item1.revision == 1
    assert item1.created_at == parser.parse("2021-02-11T13:43:22.396606")
    assert item1.size == 500

    assert item2.revision == 2
    assert item2.created_at == parser.parse("2021-02-11T14:23:55.659148")
    assert item2.size == 420


# -- tests for OCI related functions


def test_get_oci_registry_credentials(client_mock, config):
    """Get the credentials to hit the OCI Registry."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_json.return_value = {
        "image-name": "test-image-name",
        "username": "jane-doe",
        "password": "oh boy this is so secret!",
    }
    result = store.get_oci_registry_credentials("charm-name", "resource-name")

    assert client_mock.mock_calls == [
        call.request_urlpath_json(
            "GET", "/v1/charm/charm-name/resources/resource-name/oci-image/upload-credentials"
        )
    ]
    assert result.image_name == "test-image-name"
    assert result.username == "jane-doe"
    assert result.password == "oh boy this is so secret!"


def test_get_oci_image_blob(client_mock, config):
    """Get the blob generated by Charmhub to refer to the OCI image."""
    store = Store(config.charmhub)
    client_mock.request_urlpath_text.return_value = "some opaque stuff"
    result = store.get_oci_image_blob("charm-name", "resource-name", "a-very-specific-digest")

    assert client_mock.mock_calls == [
        call.request_urlpath_text(
            "POST",
            "/v1/charm/charm-name/resources/resource-name/oci-image/blob",
            json={"image-digest": "a-very-specific-digest"},
        )
    ]
    assert result == "some opaque stuff"



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import pathlib
import sys
from unittest.mock import ANY, call, patch

import pytest
from craft_cli import CraftError
from craft_parts import Action, ActionType, PartsError, Step

from charmcraft.parts import lifecycle

pytestmark = pytest.mark.skipif(sys.platform == "win32", reason="Windows not supported")


def test_partslifecycle_bad_bootstrap(tmp_path):
    fake_error = PartsError("pumba")
    with patch("craft_parts.LifecycleManager.__init__") as mock:
        mock.side_effect = fake_error
        with pytest.raises(CraftError) as cm:
            lifecycle.PartsLifecycle(
                all_parts={},
                work_dir=pathlib.Path("/some/workdir"),
                project_dir=tmp_path,
                project_name="test",
                ignore_local_sources=["*.charm"],
            )
        exc = cm.value
        assert str(exc) == "Error bootstrapping lifecycle manager: pumba"
        assert exc.__cause__ == fake_error


def test_partslifecycle_prime_dir(tmp_path):
    data = {
        "plugin": "charm",
        "source": ".",
    }

    with patch("craft_parts.LifecycleManager.refresh_packages_list"):
        lc = lifecycle.PartsLifecycle(
            all_parts={"charm": data},
            work_dir=pathlib.Path("/some/workdir"),
            project_dir=tmp_path,
            project_name="test",
            ignore_local_sources=["*.charm"],
        )
    assert lc.prime_dir == pathlib.Path("/some/workdir/prime")


def test_partslifecycle_run_new_entrypoint(tmp_path, monkeypatch):
    data = {
        "plugin": "charm",
        "source": ".",
        "charm-entrypoint": "my-entrypoint",
        "charm-python-packages": ["pkg1", "pkg2"],
        "charm-requirements": [],
    }

    # create dispatcher from previous run
    prime_dir = tmp_path / "prime"
    prime_dir.mkdir()
    dispatch = prime_dir / "dispatch"
    dispatch.write_text(
        'JUJU_DISPATCH_PATH="${JUJU_DISPATCH_PATH:-$0}" PYTHONPATH=lib:venv ./src/charm.py'
    )

    lc = lifecycle.PartsLifecycle(
        all_parts={"charm": data},
        work_dir=tmp_path,
        project_dir=tmp_path,
        project_name="test",
        ignore_local_sources=["*.charm"],
    )

    with patch("craft_parts.LifecycleManager.clean") as mock_clean:
        with patch("craft_parts.LifecycleManager.plan") as mock_plan:
            mock_plan.side_effect = SystemExit("test")
            with pytest.raises(SystemExit, match="test"):
                lc.run(Step.PRIME)

    mock_clean.assert_called_once_with(Step.BUILD, part_names=["charm"])


def test_partslifecycle_run_same_entrypoint(tmp_path, monkeypatch):
    data = {
        "plugin": "charm",
        "source": ".",
        "charm-entrypoint": "src/charm.py",
        "charm-python-packages": ["pkg1", "pkg2"],
        "charm-requirements": [],
    }

    # create dispatcher from previous run
    prime_dir = tmp_path / "prime"
    prime_dir.mkdir()
    dispatch = prime_dir / "dispatch"
    dispatch.write_text(
        'JUJU_DISPATCH_PATH="${JUJU_DISPATCH_PATH:-$0}" PYTHONPATH=lib:venv ./src/charm.py'
    )

    lc = lifecycle.PartsLifecycle(
        all_parts={"charm": data},
        work_dir=tmp_path,
        project_dir=tmp_path,
        project_name="test",
        ignore_local_sources=["*.charm"],
    )

    with patch("craft_parts.LifecycleManager.clean") as mock_clean:
        with patch("craft_parts.LifecycleManager.plan") as mock_plan:
            mock_plan.side_effect = SystemExit("test")
            with pytest.raises(SystemExit, match="test"):
                lc.run(Step.PRIME)

    mock_clean.assert_not_called()


def test_partslifecycle_run_no_previous_entrypoint(tmp_path, monkeypatch):
    data = {
        "plugin": "charm",
        "source": ".",
        "charm-entrypoint": "my-entrypoint",
        "charm-python-packages": ["pkg1", "pkg2"],
        "charm-requirements": [],
    }

    lc = lifecycle.PartsLifecycle(
        all_parts={"charm": data},
        work_dir=tmp_path,
        project_dir=tmp_path,
        project_name="test",
        ignore_local_sources=["*.charm"],
    )

    with patch("craft_parts.LifecycleManager.clean") as mock_clean:
        with patch("craft_parts.LifecycleManager.plan") as mock_plan:
            mock_plan.side_effect = SystemExit("test")
            with pytest.raises(SystemExit, match="test"):
                lc.run(Step.PRIME)

    mock_clean.assert_called_once_with(Step.BUILD, part_names=["charm"])


def test_partslifecycle_run_actions_progress(tmp_path, monkeypatch, emitter):
    data = {
        "plugin": "nil",
        "source": ".",
    }

    lc = lifecycle.PartsLifecycle(
        all_parts={"testpart": data},
        work_dir=tmp_path,
        project_dir=tmp_path,
        project_name="test",
        ignore_local_sources=[],
    )

    action1 = Action(
        part_name="testpart", step=Step.STAGE, action_type=ActionType.RUN, reason=None
    )
    action2 = Action(
        part_name="testpart", step=Step.PRIME, action_type=ActionType.RUN, reason=None
    )

    with patch("craft_parts.LifecycleManager.plan") as mock_plan:
        mock_plan.return_value = [action1, action2]
        with patch("craft_parts.executor.executor.ExecutionContext.execute") as mock_exec:
            lc.run(Step.PRIME)

    emitter.assert_progress("Running step STAGE for part 'testpart'")
    emitter.assert_progress("Running step PRIME for part 'testpart'")
    assert mock_exec.call_args_list == [
        call([action1], stdout=ANY, stderr=ANY),
        call([action2], stdout=ANY, stderr=ANY),
    ]


def test_parthelpers_get_dispatch_entrypoint(tmp_path):
    dispatch = tmp_path / "dispatch"
    dispatch.write_text(
        'JUJU_DISPATCH_PATH="${JUJU_DISPATCH_PATH:-$0}" PYTHONPATH=lib:venv ./my/entrypoint'
    )
    entrypoint = lifecycle._get_dispatch_entrypoint(tmp_path)
    assert entrypoint == "./my/entrypoint"


def test_parthelpers_get_dispatch_entrypoint_no_file(tmp_path):
    entrypoint = lifecycle._get_dispatch_entrypoint(tmp_path)
    assert entrypoint == ""
